/* This file is a front end for the simulator that emulates the
 * EST command set, which is one of the remote protocols
 * used by GDB.
 *
 * The EST command set was chosen for its simplicity.
 * An option will be available to cause the program to
 * wait for a TCP socket session so that it can be
 * used with GDB.
 */

/*
 * Commands and other strings expected by the GDB EST interface.
 *
 * At initialization:
 * x"he\r", -* Resets the prompt, and clears repeated cmds
 *
 * During debug:
 * x"go\r",				-* continue command -*
 * x"sidr\r",			-* single step -*
 * x"\003",				-* ^C interrupts the program -*
 * x"sb %x\r",			-* set a breakpoint -*
 * x"rb %x\r",			-* clear a breakpoint -*		is argument address or breakpoint ID?
 * x"rb\r",				-* clear all breakpoints -*
 * x"bfb %x %x %x\r",	-* fill (start end val) -*
 * x"smb %x %x\r",		-* setmem.cmdb (addr, value) -*
 * x"smw %x %x\r",		-* setmem.cmdw (addr, value) -*
 * x"sml %x %x\r",		-* setmem.cmdl (addr, value) -*
 * x"dmb %x %x\r",		-* getmem.cmdb (addr, len) -*
 * x"dmw %x %x\r",		-* getmem.cmdw (addr, len) -*
 * x"dml %x %x\r",		-* getmem.cmdl (addr, len) -*
 * ": ",				-* getmem.resp_delim -*
 * x"sr %s %x\r",		-* setreg.cmd (name, value) -*
 * x"dr %s\r",			-* getreg.cmd (name) -*
 * x"dr\r",				-* dump_registers -*
 * "\\(\\w+\\) = \\([0-9a-fA-F]+\\)", -* register_pattern -*
 * NULL,				-* load_routine (defaults to SRECs) -*
 * "dl\r",				-* download command -*
 * "+",					-* load response -*
 * ">BKM>",				-* monitor command prompt -*
 * "\r",				-* end-of-line terminator -*
 * "D0", "D1", "D2", "D3", "D4", "D5", "D6", "D7",
 * "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7",
 * "SR", "PC",
 */

#include <stdlib.h>
#include <string.h>
#include <strings.h>	/* For SunOS */
#include <signal.h>
#include <termios.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/socket.h>	/* For SunOS */
#include "sim68k/simulator/extern.h"
#include "sim68k/simulator/def.h"       /* include file for constants & types */
#include "sim68k/simulator/proto.h"		/* function prototypes */
#include "sim68k/devices/cif.h"

/*
 *	Local macro definitions
 */

#define HOSTNAME_LENGTH	128

/*
 *	Local type definitions
 */

enum {
	cmdLineStateHunt,
	cmdLineStateBuild
	};

typedef struct CommandRec{
	const char	*command;
	void (*exec)(const char *args,SimContext *context);
	const char	*help;
	}CommandRec;

/*
 *	Local function prototypes
 */

static const char *stripNotSpace(const char *s);
static const char *stripSpace(const char *s);
static void init(	SimContext		*context,
					FILE*			inputFileDesc,
					FILE*			outputFileDesc,
					int				port,
					unsigned long	initialPC,
					unsigned long	initialSP,
					const char*		objFileName
					);
static void reset(	SimContext		*context,
					unsigned long	startingProgramCounter,
					unsigned long	startingStackAddress,
					const char*		objFileName
					);
static void run(SimContext *context);
static void printInvalidCommand(SimContext* context);
static void printBkmPrompt(SimContext* context);
static void printRunPrompt(SimContext* context);
static void printDownloadAckPrompt(SimContext* context);
static void estSingleStepCommand(const char *args,SimContext *context);
static void estGoCommand(const char *args,SimContext *context);
static void estSetBreakPointCommand(const char *args,SimContext *context);
static void estClearBreakPointCommand(const char *args,SimContext *context);
static void estResetPromptCommand(const char *args,SimContext *context);
static void estFillCommand(const char *args,SimContext *context);
static void estSetMemoryByteCommand(const char *args,SimContext *context);
static void estSetMemoryWordCommand(const char *args,SimContext *context);
static void estSetMemoryLongCommand(const char *args,SimContext *context);
static void estDisplayMemoryByteCommand(const char *args,SimContext *context);
static void estDisplayMemoryWordCommand(const char *args,SimContext *context);
static void estDisplayMemoryLongCommand(const char *args,SimContext *context);
static void estSetRegisterCommand(const char *args,SimContext *context);
static void estDisplayRegisterCommand(const char *args,SimContext *context);
static void estDisplayAllRegisters(SimContext *context);
static void estDownloadCommand(const char *args,SimContext *context);
static void tempLoadSrecords(const char *args,SimContext *context);
static void resetCommand(const char *args,SimContext *context);
static void rwAnyDataAccessBreakCommand(const char *args,SimContext *context);
static void writeDataAccessBreakCommand(const char *args,SimContext *context);
static void readDataAccessBreakCommand(const char *args,SimContext *context);
static void displayPcTrace(const char *args,SimContext *context);
static void saveTrace(const char *args,SimContext *context);
static void commandLoop(SimContext* context);
static void prepareServer(int port,FILE** inStream,FILE** outStream);
static const char*	getCommandLine(SimContext* context,int echo);

static void interruptSignalHandler(int sig);

/*
 *	Local variables
 */

static const CommandRec	commands[]	={
 							{"go",estGoCommand,"go"},
							{"sidr",estSingleStepCommand,""},
							{"sb",estSetBreakPointCommand,""},
							{"rb",estClearBreakPointCommand,""},
							{"he",estResetPromptCommand,""},
							{"bfb",estFillCommand,""},
							{"smb",estSetMemoryByteCommand,""},
							{"smw",estSetMemoryWordCommand,""},
							{"sml",estSetMemoryLongCommand,""},
							{"dmb",estDisplayMemoryByteCommand,""},
							{"dmw",estDisplayMemoryWordCommand,""},
							{"dml",estDisplayMemoryLongCommand,""},
							{"sr",estSetRegisterCommand,""},
							{"dr",estDisplayRegisterCommand,""},
							{"dl",estDownloadCommand,""},
							{"ld",tempLoadSrecords,""},
							{"reset",resetCommand,"[address]"},
							{"dbrw",rwAnyDataAccessBreakCommand,"<address> <name>"},
							{"dbw",writeDataAccessBreakCommand,"<address> <data> <name>"},
							{"dbr",readDataAccessBreakCommand,"<address> <data> <name>"},
							{"dt",displayPcTrace,"<starting_position> (always displays 50)"},
							{"st",saveTrace,"<filename> (save whole trace to file)"},
							{(const char *)0,(void(*)(const char*,SimContext*))0,(const char *)0}
							};

static int showRegisterDumps = 0;

/*
 *	Global Variables
 */

int					SIGINT_Received;
static SimContext	context;

static void	printHelp(void)
{
	printf(	"Usage: sim68k [options]\n"
			" -h print this help text.\n"
			" -f specify name of the object file to be loaded.\n"
			" -r show register dumps\n"
			" -p specify port number for inet server\n"
			" -i specify initial PC on reset.\n"
			" -s specify initial SP on reset.\n"
			);
}

int main(int argc,char *argv[]){
struct sigaction	act;
struct sigaction	oact;

	int opt;
	const char			*objFileName	= (const char *)0;
	int					port		= 9600;	/* default IP port */
	FILE*				inStream	= stdin;
	FILE*				outStream	= stdout;
	unsigned long		initialPC	= 0x0400;
	unsigned long		initialSP	= 0x0F00;

	SIGINT_Received	= FALSE;

	act.sa_handler	= interruptSignalHandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags	= 0;
	sigaction(SIGINT,&act,&oact);

	while((opt = getopt(argc,argv,"hi:s:f:p:r")) != -1){
		switch(opt){
			case 'h':
				printHelp();
				exit(1);
			case 'f':
				objFileName	= optarg;
				break;
			case 'r':
				showRegisterDumps = 1;
				break;
			case 'p':
				port	= (int)strtol(optarg,(char **)0,0);
				prepareServer(port,&inStream,&outStream);
				break;
			case 'i':	/* Initial (RESET) PC */
				initialPC	= strtoul(optarg,(char **)0,0);
				break;
			case 's':	/* Initial (RESET) SP */
				initialSP	= strtoul(optarg,(char **)0,0);
				break;
			case ':':
				fprintf(stderr,"option %c requires an argument\n",optopt);
				exit(1);
			case '?':
				fprintf(stderr,"unknown option: %c\n",optopt);
				exit(1);
			}
		}

	init(&context,inStream,outStream,port,initialPC,initialSP,objFileName);
	commandLoop(&context);
	return 0;
	}

static void prepareServer(int port,FILE** inStream,FILE** outStream){
	int					server_sockfd;
	int					client_sockfd;
	int					server_len;
	int					client_len;
	struct sockaddr_in	server_address;
	struct sockaddr_in	client_address;

	/*	Create a socket for the server */

	server_sockfd	= socket(AF_INET,SOCK_STREAM,0);

	/* Name the socket */

	server_address.sin_family	= AF_INET;
	server_address.sin_addr.s_addr		= htonl(INADDR_ANY);
	server_address.sin_port		= htons(port);
	server_len	= sizeof(server_address);
	bind(server_sockfd,(struct sockaddr *)&server_address, server_len);

	listen(server_sockfd,1);
#ifdef DEBUG
	fprintf(stderr,"server waiting\n");
#endif
	client_sockfd	= accept(server_sockfd,(struct sockaddr *)&client_address,&client_len);
	*inStream		= fdopen(client_sockfd,"r");
	if(!*inStream){
		fprintf(stderr,"cant open input stream!\n");
		exit(1);
		}
	*outStream		= fdopen(client_sockfd,"w");
	if(!*outStream){
		fprintf(stderr,"cant open output stream!\n");
		exit(1);
		}
	}

static void commandLoop(SimContext* context){
	void								(*lastCommand)(const char*,SimContext*) = 0;
	char								*b = 0;
	const char							*prev = 0;
	const char							*line;
	const char							*cmd;
	int									i;

	printBkmPrompt(context);
	while((line = getCommandLine(context,1)) && !SIGINT_Received){
		if(strlen(line)){
			for(i=0;(cmd=commands[i].command);i++){
				if(!strncmp(cmd,line,strlen(cmd))){
					line	= (char *)stripNotSpace(line);
					line	= (char *)stripSpace(line);
#ifdef DEBUG
					fprintf(stderr,"Command: %s %s\n",commands[i].command,line);
#endif
					(*commands[i].exec)((char *)line,context);
					lastCommand	= commands[i].exec;
					prev		= b;
					break;
					}
				}
			if(!cmd){
				printInvalidCommand(context);
				}
			}
		else{
			if(lastCommand){
				line	= (char *)stripSpace(prev);
				line	= (char *)stripNotSpace(line);
				line	= (char *)stripSpace(line);
				(*lastCommand)((char *)line,context);
				}
			}
		printBkmPrompt(context);
		}
	fclose(context->inputFileDesc);
	fclose(context->outputFileDesc);
	closeDevices();
	exit(0);
	}

static void printInvalidCommand(SimContext* context){
	fprintf(context->outputFileDesc,"!Error! invalid command\n");
	fflush(context->outputFileDesc);
	}

static void printBkmPrompt(SimContext* context){
	fprintf(context->outputFileDesc,"\n>BKM>");
	fflush(context->outputFileDesc);
	}

static void printRunPrompt(SimContext* context){
	fprintf(context->outputFileDesc,"\n>RUN>");
	fflush(context->outputFileDesc);
	}

static void printDownloadAckPrompt(SimContext* context){
	fprintf(context->outputFileDesc,"+");
	fflush(context->outputFileDesc);
	}

static void estGoCommand(const char *args,SimContext *context){
	context->SSTEP = FALSE;
	printRunPrompt(context);
	run(context);
	}

static void estSingleStepCommand(const char *args,SimContext *context){
	context->SSTEP = TRUE;
	run(context);
	estDisplayAllRegisters(context);
	}

static void estSetBreakPointCommand(const char *args,SimContext *context){
	int				i;
	unsigned long	loc;

	loc	= strtoul(args,(char **)0,16);

	/* set breakpoint in table and increment counter */

	for (i = 0; i<context->bpoints; i++)
		if(context->brkpt[i] == loc)
			return;
	context->brkpt[context->bpoints++] = loc;
	if (context->bpoints > 100)
		context->bpoints = 100;
	}

static void estClearBreakPointCommand(const char *args,SimContext *context){
	int				i;
	int				j;
	unsigned long	loc;

	if(!strlen(args)){
		/* clear all break points */
		context->bpoints	= 0;
		return;
		}

	loc	= strtoul(args,(char **)0,16);

	/* find break point specified */

	for (i = 0; i < context->bpoints; i++) if (context->brkpt[i] == loc) break;

	if (i == context->bpoints)	/* if not found, print message */
		{
		return;
		}
	--context->bpoints;
	for (j = i; j < context->bpoints; j++) /* adjust breakpoint table and */
		context->brkpt[j] = context->brkpt[j + 1];	/* decrement counter */
	}

static void estResetPromptCommand(const char *args,SimContext *context){
	int	i;
	for(i=0;commands[i].command;i++){
		fprintf(context->outputFileDesc,"%s %s\n",commands[i].command,commands[i].help);
		}
	}

static void estFillCommand(const char *args,SimContext *context){
	unsigned long	start;
	unsigned long	end;
	unsigned long	val;

	start	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	end		= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	val		= strtoul(args,(char **)0,16);

	for(;start<=end;start++){
		context->memory[start % MEMSIZE] = (unsigned char) val;
		}
	}

static void estSetMemoryByteCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	value;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	value	= strtoul(args,(char **)0,16);

	context->memory[addr % MEMSIZE] = (unsigned char) value;
	}

static void estSetMemoryWordCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	value;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	value	= strtoul(args,(char **)0,16);

	context->memory[addr % MEMSIZE] = (unsigned char) (value>>8);
	context->memory[(addr+1) % MEMSIZE] = (unsigned char) (value);
	}

static void estSetMemoryLongCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	value;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	value	= strtoul(args,(char **)0,16);

	context->memory[addr % MEMSIZE]		= (unsigned char) (value>>24);
	context->memory[(addr+1) % MEMSIZE]	= (unsigned char) (value>>16);
	context->memory[(addr+2) % MEMSIZE]	= (unsigned char) (value>>8);
	context->memory[(addr+3) % MEMSIZE]	= (unsigned char) (value);
	}

static void estDisplayMemoryByteCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	beyondEnd;
	unsigned long	p;
	unsigned long	len;
	unsigned long	i;
	unsigned char	c;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	len		= strtoul(args,(char **)0,16);

	beyondEnd		= addr+len;

	for(;addr<beyondEnd;addr+=16){
		fprintf(context->outputFileDesc,"  %8.8lX:",addr);
		for(i=0,p=addr;i<16;i++,p++){
			if(p<beyondEnd){
				fprintf(context->outputFileDesc," %2.2X",context->memory[p % MEMSIZE]);
				}
			else{
				fprintf(context->outputFileDesc,"   ");
				}
			}
		fprintf(context->outputFileDesc,"   ");
		for(i=0,p=addr;i<16;i++,p++){
			if(p<beyondEnd){
				c	= context->memory[p % MEMSIZE];
				if(isprint(c) && !iscntrl(c)){
					fprintf(context->outputFileDesc,"%c",c);
					}
				else{
					fprintf(context->outputFileDesc,"%c",'.');
					}
				}
			else{
				break;
				}
			}
		fprintf(context->outputFileDesc,"\n");
		}

	fflush(context->outputFileDesc);
	}

static void estDisplayMemoryWordCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	beyondEnd;
	unsigned long	p;
	unsigned long	len;
	unsigned long	i;
	unsigned char	c;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	len		= strtoul(args,(char **)0,16);

	beyondEnd		= addr+2*len;

	for(;addr<beyondEnd;addr+=16){
		fprintf(context->outputFileDesc,"  %8.8lX:",addr);
		for(i=0,p=addr;i<8;i++,p+=2){
			if(p<beyondEnd){
				fprintf(context->outputFileDesc," %2.2X",context->memory[p % MEMSIZE]);
				fprintf(context->outputFileDesc,"%2.2X",context->memory[p+1 % MEMSIZE]);
				}
			else{
				fprintf(context->outputFileDesc,"     ");
				}
			}
		fprintf(context->outputFileDesc,"   ");
		for(i=0,p=addr;i<16;i++,p++){
			if(p<beyondEnd){
				c	= context->memory[p % MEMSIZE];
				if(isprint(c) && !iscntrl(c)){
					fprintf(context->outputFileDesc,"%c",c);
					}
				else{
					fprintf(context->outputFileDesc,"%c",'.');
					}
				}
			else{
				break;
				}
			}
		fprintf(context->outputFileDesc,"\n");
		}

	fflush(context->outputFileDesc);
	}

static void estDisplayMemoryLongCommand(const char *args,SimContext *context){
	unsigned long	addr;
	unsigned long	beyondEnd;
	unsigned long	p;
	unsigned long	len;
	unsigned long	i;
	unsigned char	c;

	addr	= strtoul(args,(char **)0,16);
	args	= stripNotSpace(args);
	args	= stripSpace(args);
	len		= strtoul(args,(char **)0,16);

	beyondEnd		= addr+4*len;

	for(;addr<beyondEnd;addr+=16){
		fprintf(context->outputFileDesc,"  %8.8lX:",addr);
		for(i=0,p=addr;i<4;i++,p+=4){
			if(p<beyondEnd){
				fprintf(context->outputFileDesc," %2.2X",context->memory[p % MEMSIZE]);
				fprintf(context->outputFileDesc,"%2.2X",context->memory[p+1 % MEMSIZE]);
				fprintf(context->outputFileDesc,"%2.2X",context->memory[p+2 % MEMSIZE]);
				fprintf(context->outputFileDesc,"%2.2X",context->memory[p+3 % MEMSIZE]);
				}
			else{
				fprintf(context->outputFileDesc,"         ");
				}
			}
		fprintf(context->outputFileDesc,"   ");
		for(i=0,p=addr;i<16;i++,p++){
			if(p<beyondEnd){
				c	= context->memory[p % MEMSIZE];
				if(isprint(c) && !iscntrl(c)){
					fprintf(context->outputFileDesc,"%c",c);
					}
				else{
					fprintf(context->outputFileDesc,"%c",'.');
					}
				}
			else{
				break;
				}
			}
		fprintf(context->outputFileDesc,"\n");
		}

	fflush(context->outputFileDesc);
	}

static void estSetRegisterCommand(const char *args,SimContext *context){
	const char		*valStr;
	unsigned long	value;
	unsigned		regno;

	regno	= args[1] - '0';

	valStr	= stripNotSpace(args);
	valStr	= stripSpace(valStr);
	value	= strtoul(valStr,(char **)0,16);

	switch(toupper(args[0])){
		case 'D':
			if(regno<8){
				context->D[regno]	= (long)value;
				}
			else break;
			return;
		case 'A':
			if(regno == 7){
				if(context->sr & 0x2000){
					context->A[8]	= (long)value;
					}
				else{
					context->A[7]	= (long)value;
					}
				}
			else if(regno<7){
				context->A[regno]	= (long)value;
				}
			else break;
			return;
		case 'S':
			if(toupper(args[1])=='R'){
				context->sr	= (short)value;
				}
			else break;
			return;
		case 'P':
			if(toupper(args[1])=='C'){
				context->pc	= (long)value;
				}
			else break;
			return;
		}
	}

static void estDisplayRegisterCommand(const char *args,SimContext *context){
	unsigned	regno;

	if(!strlen(args)){
		estDisplayAllRegisters(context);
		return;
		}

	regno	= args[1] - '0';

	switch(toupper(args[0])){
		case 'D':
			if(regno<8){
				fprintf(context->outputFileDesc,"D%d = %8.8lX\n",regno,context->D[regno]);
				}
			else break;
			return;
		case 'A':
			if(regno == 7){
				if(context->sr & 0x2000){
					fprintf(context->outputFileDesc,"A%d = %8.8lX\n",regno,context->A[8]);
					}
				else{
					fprintf(context->outputFileDesc,"A%d = %8.8lX\n",regno,context->A[7]);
					}
				}
			else if(regno<8){
				fprintf(context->outputFileDesc,"A%d = %8.8lX\n",regno,context->A[regno]);
				}
			else break;
			return;
		case 'S':
			if(toupper(args[1])=='R'){
				fprintf(context->outputFileDesc,"SR = %4.4X\n",(unsigned)(context->sr));
				}
			else break;
			return;
		case 'P':
			if(toupper(args[1])=='C'){
				fprintf(context->outputFileDesc,"PC = %8.8lX\n",context->pc);
				}
			else break;
			return;
		}
	}

static void estDisplayAllRegisters(SimContext *context){
	int	i;
	int	j;
	int	k;
	int	l;

	for(i=0;i<2;i++){
		for(j=0;j<4;j++){
			k	= i*4+j;
			fprintf(context->outputFileDesc,"D%d = %8.8lX ",k,context->D[k]);
			if(showRegisterDumps)
				printf("D%d = %8.8lX ",k,context->D[k]);
			}
		fprintf(context->outputFileDesc,"\n");
		if(showRegisterDumps)
			printf("\n");
		}

	for(i=0;i<2;i++){
		for(j=0;j<4;j++){
			k	= i*4+j;
			l	= ((k==7) && (context->sr & 0x2000))?8:k;
			fprintf(context->outputFileDesc,"A%d = %8.8lX ",k,context->A[l]);
			if(showRegisterDumps)
				printf("A%d = %8.8lX ",k,context->A[l]);
			}
		fprintf(context->outputFileDesc,"\n");
		if(showRegisterDumps)
			printf("\n");
		}

	if(context->sr & 0x2000){
		fprintf(context->outputFileDesc,"USP = %8.8lX ",context->A[7]);
		if(showRegisterDumps)
			printf("USP = %8.8lX ",context->A[7]);
		fprintf(context->outputFileDesc,"SSP*= %8.8lX ",context->A[8]);
		if(showRegisterDumps)
			printf("SSP*= %8.8lX ",context->A[8]);
		}
	else{
		fprintf(context->outputFileDesc,"USP*= %8.8lX ",context->A[7]);
		if(showRegisterDumps)
			printf("USP*= %8.8lX ",context->A[7]);
		fprintf(context->outputFileDesc,"SSP = %8.8lX ",context->A[8]);
		if(showRegisterDumps)
			printf("SSP = %8.8lX ",context->A[8]);
		}
	fprintf(context->outputFileDesc,"PC = %8.8lX ",context->pc);
	if(showRegisterDumps)
		printf("PC = %8.8lX ",context->pc);
	fprintf(context->outputFileDesc,"SR = %4.4lX\n",context->sr);
	if(showRegisterDumps)
		printf("SR = %4.4lX\n",context->sr);
	}

static void tempLoadSrecords(const char *args,SimContext *context){
	/* This function is only for testing under stdin & stdout */
	mfill(args,context);
	}

static void estDownloadCommand(const char *args,SimContext *context){
	const char*	line;
	printDownloadAckPrompt(context);
	while((line=getCommandLine(context,0))){
		if(line[0] != 'S') return;
		if((line[1] == '8') || (line[1] == '9')){
			return;
			}
		else{
			printf("%s\n",line);
			printDownloadAckPrompt(context);
			}
		}
	}

static void resetCommand(const char *args,SimContext *context){
	const char*		addrStr;
	const char*		stackStr;
	unsigned long	startingProgramCounter;
	unsigned long	startingStackAddress;

	addrStr		= stripSpace(args);
	stackStr	= stripNotSpace(addrStr);
	stackStr	= stripSpace(stackStr);

	if(strlen(addrStr)){
		startingProgramCounter	= strtoul(addrStr,(char **)0,0);
		}
	else{
		startingProgramCounter	= context->initialPC;
		}
	if(strlen(stackStr)){
		startingStackAddress	= strtoul(stackStr,(char **)0,0);
		}
	else{
		startingStackAddress	= context->initialSP;
		}
	reset(context,startingProgramCounter,startingStackAddress,context->objFileName);
	}

static void writeDataAccessBreakCommand(const char *args,SimContext *context){
	SimAccessBreakpoint*	b;
	const char*		name;
	const char*		addrStr;
	const char*		dataStr;
	unsigned long	data;
	unsigned long	address;

	addrStr		= stripSpace(args);
	dataStr		= stripNotSpace(addrStr);
	dataStr		= stripSpace(dataStr);
	name		= stripNotSpace(dataStr);
	name		= stripSpace(name);

	if( (!strlen(addrStr)) || (!strlen(dataStr)) || (!strlen(name)) ){
		fprintf(context->outputFileDesc,"Command error:\n");
		fprintf(context->outputFileDesc,"dbw <address> <data> <name>\n");
		return;
		}

	address		= strtoul(addrStr,(char **)0,0);
	data		= strtoul(dataStr,(char **)0,0);

	b	= newSimAccessBreakpoint(name,address,accessModeWrite,(unsigned char)data,(unsigned char)~0);
	if(b){
		fprintf(context->outputFileDesc,"New Data Breakpoint:\n");
		fprintf(context->outputFileDesc,"  name: %s\n",b->name);
		fprintf(context->outputFileDesc,"  address: 0x%4.4lX\n",b->address);
		fprintf(context->outputFileDesc,"  data: 0x%2.2X\n",b->data);
		putAccessBreakpoint(context,b);
		}
	else{
		fprintf(context->outputFileDesc,"Data access breakpoint allocation failed\n");
		}
	}

static void readDataAccessBreakCommand(const char *args,SimContext *context){
	SimAccessBreakpoint*	b;
	const char*		name;
	const char*		addrStr;
	const char*		dataStr;
	unsigned long	data;
	unsigned long	address;

	addrStr		= stripSpace(args);
	dataStr		= stripNotSpace(addrStr);
	dataStr		= stripSpace(dataStr);
	name		= stripNotSpace(dataStr);
	name		= stripSpace(name);

	if( (!strlen(addrStr)) || (!strlen(dataStr)) || (!strlen(name)) ){
		fprintf(context->outputFileDesc,"Command error:\n");
		fprintf(context->outputFileDesc,"dbr <address> <data> <name>\n");
		return;
		}

	address		= strtoul(addrStr,(char **)0,0);
	data		= strtoul(dataStr,(char **)0,0);

	b	= newSimAccessBreakpoint(name,address,accessModeRead,(unsigned char)data,(unsigned char)~0);
	if(b){
		fprintf(context->outputFileDesc,"New Data Breakpoint:\n");
		fprintf(context->outputFileDesc,"  name: %s\n",b->name);
		fprintf(context->outputFileDesc,"  address: 0x%4.4lX\n",b->address);
		fprintf(context->outputFileDesc,"  data: 0x%2.2X\n",b->data);
		putAccessBreakpoint(context,b);
		}
	else{
		fprintf(context->outputFileDesc,"Data access breakpoint allocation failed\n");
		}
	}

static void rwAnyDataAccessBreakCommand(const char *args,SimContext *context){
	SimAccessBreakpoint*	b;
	const char*		name;
	const char*		addrStr;
	unsigned long	address;

	addrStr		= stripSpace(args);
	name		= stripNotSpace(addrStr);
	name		= stripSpace(name);

	if( (!strlen(addrStr)) || (!strlen(name)) ){
		fprintf(context->outputFileDesc,"Command error:\n");
		fprintf(context->outputFileDesc,"dbrw <address> <name>\n");
		return;
		}
	address		= strtoul(addrStr,(char **)0,0);

	b	= newSimAccessBreakpoint(name,address,accessModeAny,(unsigned char)0,(unsigned char)~0);
	if(b){
		printf("New Data Breakpoint:\n");
		printf("  name: %s\n",b->name);
		printf("  address: 0x%4.4lX\n",b->address);
		putAccessBreakpoint(context,b);
		}
	else{
		fprintf(context->outputFileDesc,"Data access breakpoint allocation failed\n");
		}
	}

static void displayPcTrace(const char *args,SimContext *context){
	unsigned long	count;
	unsigned long	nDisplay	= 50;
	PCTraceRecord*	next;
	const char*		startStr;
	unsigned long	start;

	for(count=0,next=context->head;next;count++,next=next->link);

	startStr		= stripSpace(args);
	if(strlen(startStr)){
		start		= strtoul(startStr,(char **)0,0);
		}
	else{
		start		= nDisplay;
		}

	start	= (start > count)?count:start;

	if(nDisplay < count){
		for(count=count-start,next=context->head;count;count--,next=next->link);
		}
	else{
		nDisplay	= count;
		}

	for(start--;next && nDisplay;next=next->link,nDisplay--,start--){
		switch(next->type){
			case pcTraceRecordTypeInstruction:
				fprintf(context->outputFileDesc,"%8.8lu Opcode 0x%8.8lX\n",start,next->address);
				break;
			case pcTraceRecordTypeRead:
				switch(next->size){
					case pcTraceRecordSizeByte:
						fprintf(context->outputFileDesc,"%8.8lu Read   0x%8.8lX 0x%2.2lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeWord:
						fprintf(context->outputFileDesc,"%8.8lu Read   0x%8.8lX 0x%4.4lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeLong:
						fprintf(context->outputFileDesc,"%8.8lu Read   0x%8.8lX 0x%8.8lx\n",start,next->address,next->data);
						break;
					}
				break;
			case pcTraceRecordTypeWrite:
				fprintf(context->outputFileDesc,"%8.8lu Write  0x%8.8lX\n",start,next->address);
				switch(next->size){
					case pcTraceRecordSizeByte:
						fprintf(context->outputFileDesc,"%8.8lu Write  0x%8.8lX 0x%2.2lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeWord:
						fprintf(context->outputFileDesc,"%8.8lu Write  0x%8.8lX 0x%4.4lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeLong:
						fprintf(context->outputFileDesc,"%8.8lu Write  0x%8.8lX 0x%8.8lx\n",start,next->address,next->data);
						break;
					}
				break;
			}
		}
	}

static void saveTrace(const char *args,SimContext *context){
	PCTraceRecord*	next;
	const char*		filename;
	unsigned long	start;
	FILE			*file;

	filename		= stripSpace(args);
	if(strlen(filename)){
		file	= fopen(filename,"w");
		if(!file){
			fprintf(context->outputFileDesc,"Can't open file: %s\n",filename);
			return;
			}
		}
	else{
		fprintf(context->outputFileDesc,"Requires file name.\n");
		return;
		}

	for(start=0,next=context->head;next;start++,next=next->link){
		switch(next->type){
			case pcTraceRecordTypeInstruction:
				fprintf(file,"%8.8lu Opcode 0x%8.8lX\n",start,next->address);
				break;
			case pcTraceRecordTypeRead:
				switch(next->size){
					case pcTraceRecordSizeByte:
						fprintf(file,"%8.8lu Read   0x%8.8lX 0x%2.2lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeWord:
						fprintf(file,"%8.8lu Read   0x%8.8lX 0x%4.4lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeLong:
						fprintf(file,"%8.8lu Read   0x%8.8lX 0x%8.8lx\n",start,next->address,next->data);
						break;
					}
				break;
			case pcTraceRecordTypeWrite:
				fprintf(file,"%8.8lu Write  0x%8.8lX\n",start,next->address);
				switch(next->size){
					case pcTraceRecordSizeByte:
						fprintf(file,"%8.8lu Write  0x%8.8lX 0x%2.2lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeWord:
						fprintf(file,"%8.8lu Write  0x%8.8lX 0x%4.4lx\n",start,next->address,next->data);
						break;
					case pcTraceRecordSizeLong:
						fprintf(file,"%8.8lu Write  0x%8.8lX 0x%8.8lx\n",start,next->address,next->data);
						break;
					}
				break;
			}
		}
	fclose(file);
	}

static void run(SimContext *context){
	int				err;
	int				save=0;
	struct termios	savedTios;
	struct termios	newTios;

	err	= tcgetattr(fileno(context->inputFileDesc),&savedTios);
	if(err){
#ifdef DEBUG
		fprintf(stderr,"tcgetattr() failed\n");
#endif
		}
	newTios	= savedTios;
	newTios.c_lflag	&= ~ICANON;
	newTios.c_lflag	&= ~ECHO;
	newTios.c_cc[VMIN]	= 0;
	newTios.c_cc[VTIME]	= 0;
	err	= tcsetattr(fileno(context->inputFileDesc),TCSANOW,&newTios);
	if(err){
#ifdef DEBUG
		fprintf(stderr,"tcsetattr() connonical off failed\n");
#endif
		save	= fcntl(fileno(context->inputFileDesc),F_GETFL);
		if(save == -1){
			fprintf(stderr,"fcntl() O_NONBLOCK failed\n");
			}
		else{
			save	= fcntl(fileno(context->inputFileDesc),F_SETFL,(long)O_NONBLOCK);
			}
		}

	flushTrace(context);

	runprog(context);

	err	= tcsetattr(fileno(context->inputFileDesc),TCSANOW,&savedTios);
	if(err){
#ifdef DEBUG
		fprintf(stderr,"tcsetattr() restore failed\n");
#endif
		if(save != -1){
			save	= fcntl(fileno(context->inputFileDesc),F_SETFL,(long)save);
			}
		}
	context->old_pc = context->pc;
	}

static void reset(	SimContext		*context,
					unsigned long	startingProgramCounter,
					unsigned long	startingStackAddress,
					const char*		objFileName
					)
{
int	i;
	resetDevices();
	for (i = 0; i <= 7; i++)
		context->A[i] = context->D[i] = context->old_a[i] = context->old_d[i] = 0;
	context->old_a[8]				= 0;
	context->cycles					= 0;
	context->sr						= 0x2700;
	context->pc						= startingProgramCounter;
	context->A[8]					= startingStackAddress;
	context->old_pc					= -1;
	context->old_CYCLES 			= -1;
	context->old_sr					= -1; /* set different from 'pc' and 'cycles' */
	context->old_pc					= -1;
	context->old_CYCLES				= -1;
	context->old_sr					= -1; /* set different from 'pc' and 'cycles' */
	context->busOrAddressErrorInProgress	= 0;
	context->doubleBusFault			= 0;
	if(objFileName) tempLoadSrecords(objFileName,context);
}

static void init(	SimContext		*context,
					FILE*			inputFileDesc,
					FILE*			outputFileDesc,
					int				port,
					unsigned long	initialPC,
					unsigned long	initialSP,
					const char*		objFileName
					)    /* initialization for the simulator */
{
	int	i;
	PCTraceRecord*	t;

	context->inputFileDesc		= inputFileDesc;
	context->outputFileDesc		= outputFileDesc;
	initializeDevices(inputFileDesc,port);
	context->exceptions = TRUE;
	reset(context,initialPC,initialSP,objFileName);
	context->lbuf[0] = '\0';		/* initialize to prevent memory access violations */
	context->wordptr[0] = context->lbuf;
	context->TRACE = context->SSTEP = FALSE;
	context->old_TRACE = context->old_SSTEP = TRUE;	  /* set different from 'TRACE' and 'SSTEP' */
	context->dataBreakPoint			= FALSE;
	context->writeDataBreakList		= 0;
	context->writeAnyBreakList		= 0;
	context->readDataBreakList		= 0;
	context->readAnyBreakList		= 0;
	context->rwAnyBreakList			= 0;
	context->head					= 0;
	context->tail					= 0;
	context->pool					= 0;
	context->busOrAddressErrorInProgress	= 0;
	context->doubleBusFault			= 0;
	context->initialPC				= initialPC;
	context->initialSP				= initialSP;
	context->objFileName			= objFileName;
	context->bpoints				= 0;
	for(i=0;i<(64*1024);i++){
		t	= (PCTraceRecord*)malloc(sizeof(PCTraceRecord));
		if(!t) break;
		freeTraceRec(context,t);
		}
}

static const char *stripSpace(const char *s){
	while(*s && isspace(*s))	s++;
	return s;
	}

static const char *stripNotSpace(const char *s){
	while(*s && !(isspace(*s)))	s++;
	return s;
	}

static const char*	getCommandLine(SimContext* context,int echo){
	char								ch;
	static char							buff1[1024];
	static char							buff2[1024];
	static char							*prev;
	static char							terminator	= '\0';
	int									i = 0;
	int									state;
	char								*b;

	b	= (char *)(prev==buff1)?buff2:buff1;

	for(state=cmdLineStateHunt;(0 < read(fileno(context->inputFileDesc),&ch,1)) & !SIGINT_Received;){
#ifdef DEBUG
		fprintf(stderr,"	%c	%2.2X\n",ch,(int)ch);
#endif
		if(echo){
			if(0 > write(fileno(context->outputFileDesc),&ch,1)){
				fclose(context->inputFileDesc);
				fclose(context->outputFileDesc);
				exit(0);
				}
			}
		switch(state){
			case cmdLineStateHunt:
				if((ch == '\n') || (ch == '\r')){
					if(ch == terminator){
						if(prev){
							return prev;
							}
						}
					}
				if(isalpha(ch)){
					i			= 0;
					b[i]		= ch;
					i++;
					state		= cmdLineStateBuild;
					}
				break;
			case cmdLineStateBuild:
				if((ch == '\n') || (ch == '\r')){
					terminator	= ch;
					b[i]		= '\0';
					prev		= b;
					return b;
					}
				if(i < 1023){
					b[i]	= ch;
					i++;
					}
				break;
			default:
				fprintf(stderr,"illegal command state\n");
				exit(1);
			}
		}
	return (char *)0;
	}

static void interruptSignalHandler(int sig){
	SIGINT_Received	= TRUE;
	}
