#include "m2681sim.h"
#include "m2681tx.h"

using namespace Oscl;

M2681Tx::M2681Tx(M2681TxOwner& owner):
					_owner(owner),
					_resetState(*this),
					_txFifo1(*this),
					_txFifo2(*this),
					_txFifo3(*this),
					_txFifo4(*this)
					{
	_state			= &_resetState;
	_rx				= 0;
	_currentFifo	= 0;
	_activeFifo		= 0;
	_enabled		= false;
	_emptyList.put(&_txFifo1);
	_emptyList.put(&_txFifo2);
	_emptyList.put(&_txFifo3);
	_emptyList.put(&_txFifo4);
	_enabled		= true;
	}

M2681Tx::M2681Tx(M2681TxOwner& owner,OctetReceiver* rx):
						_rx(rx),
						_owner(owner),
						_resetState(*this),
						_txFifo1(*this,rx),
						_txFifo2(*this,rx),
						_txFifo3(*this,rx),
						_txFifo4(*this,rx)
						{
	_state			= &_resetState;
	_currentFifo	= 0;
	_activeFifo		= 0;
	_enabled		= false;
	_emptyList.put(&_txFifo1);
	_emptyList.put(&_txFifo2);
	_emptyList.put(&_txFifo3);
	_emptyList.put(&_txFifo4);
	_enabled		= true;
	changeReceiver(_rx);
	}

void M2681Tx::reset(){
	M2681TxFifo*	savedCurrentFifo	= _currentFifo;
	M2681TxFifo*	savedActiveFifo		= _activeFifo;
	M2681TxFifo*	next;
	_enabled		= false;
	_currentFifo	= 0;
	_activeFifo		= 0;
	if(savedCurrentFifo) _activeList.put(savedCurrentFifo);
	if(savedActiveFifo) _activeList.put(savedActiveFifo);
	while((next=_emptyList.get())){
		_activeList.put(next);
		}
	while((next=_activeList.get())){
		next->reset();
		}
	_state				= &_resetState;
	_state->reset();
	_enabled			= true;
	}

void M2681Tx::clock(unsigned long bitClocksSinceLastClock){
	unsigned long	byteClocks;
	bitClocksSinceLastClock				+= _remainingClocksBeforeNextByteClock;
	byteClocks							= bitClocksSinceLastClock/8;
	_remainingClocksBeforeNextByteClock	= bitClocksSinceLastClock%8;
	for(;byteClocks;byteClocks--){
		if(_activeFifo) _activeFifo->byteClock();
		}
	}

void M2681Tx::enable(){
	if(!_enabled){
		_enabled			= true;
		_currentFifo		= _emptyList.get();
		if(_currentFifo)	_owner.transmitFifoEmpty();
		else				_owner.transmitFifoNotEmpty();
		}
	}

void M2681Tx::disable(){
	M2681TxFifo*	next;
	_enabled	= false;
	if(_currentFifo){
		_currentFifo->reset();
		_currentFifo	= 0;
		}
	if(_activeFifo){
		_activeFifo->reset();
		_activeFifo	= 0;
		}
	while((next=_activeList.get())){
		next->reset();
		}
	}

void M2681Tx::writeByte(unsigned char data){
	if(_currentFifo){
		_currentFifo->put((Oscl::Octet)data);
		if(!_activeFifo){
			_activeFifo	= _currentFifo;
			}
		else{
			_activeList.put(_currentFifo);
			}
		_currentFifo		= _emptyList.get();
		if(_currentFifo)	_owner.transmitFifoEmpty();
		else				_owner.transmitFifoNotEmpty();
		}
	}

OctetReceiver* M2681Tx::changeReceiver(OctetReceiver* rx){
	_txFifo1.changeReceiver(rx);
	_txFifo2.changeReceiver(rx);
	_txFifo3.changeReceiver(rx);
	return _txFifo4.changeReceiver(rx);
	}

// M2681TxFifoOwner Interface

void M2681Tx::txFifoEmpty(M2681TxFifo* theFifo){
	if(!_currentFifo){
		_currentFifo	= theFifo;
		_owner.transmitFifoEmpty();
		}
	else{
		_emptyList.put(theFifo);
		}
	if(_enabled){
		_activeFifo		= _activeList.get();
		}
	}

/////////////////////////// M2681TxState ///////////////

M2681TxState::M2681TxState(M2681Tx& tx):_tx(tx){
	}

/////////////////////////// M2681TxStateReset ///////////////

M2681TxStateReset::M2681TxStateReset(M2681Tx& tx):M2681TxState(tx){
			}

void M2681TxStateReset::reset(){
	// no change in state
	}

void M2681TxStateReset::poll(unsigned long cyclesSinceLastPoll){
	// nothing to do
	}

void M2681TxStateReset::enable(){
	//_tx.state	= &_tx.enableState;
	//_tx.enableState->reset();
	}

void M2681TxStateReset::disable(){
	//_tx.state	= &_tx.disableState;
	}

// Transmit FIFO

M2681TxFifo::M2681TxFifo(M2681TxFifoOwner& owner,OctetReceiver* receiver):
								_resetState(*this),_owner(owner),_receiver(receiver){
	_state	= &_resetState;
	};

M2681TxFifo::M2681TxFifo(M2681TxFifoOwner& owner):
								_resetState(*this),_owner(owner),_receiver(0){
	_state	= &_resetState;
	};

OctetReceiver* M2681TxFifo::changeReceiver(OctetReceiver* receiver){
	OctetReceiver*	rx;
	rx			= _receiver;
	_receiver	= receiver;
	return rx;
	}

void M2681TxFifo::reset(){
	_remainingClocksBeforeNextTransmit	=0;
	_octetsToTransmit		= 0;
	_currentOctet			= 0;
	_owner.txFifoEmpty(this);
	}

void M2681TxFifo::byteClock(){
	if(_octetsToTransmit){
		if(_receiver) _receiver->receive(_fifo[_currentOctet]);
		_currentOctet++;
		_octetsToTransmit--;
		}
	else{
		_currentOctet	= 0;
		_owner.txFifoEmpty(this);
		}
	}

void M2681TxFifo::put(Oscl::Octet data){
	if(_octetsToTransmit<_fifoSize){
		_fifo[_octetsToTransmit]	= data;
		_octetsToTransmit++;
		}
	}

// M2681TxFifoState
M2681TxFifoState::M2681TxFifoState(M2681TxFifo& fifo):_fifo(fifo){
	}

// M2681TxFifoStateReset

M2681TxFifoStateReset::M2681TxFifoStateReset(M2681TxFifo& fifo):M2681TxFifoState(fifo){
	}

void M2681TxFifoStateReset::clock(unsigned long bitClocksSinceLastClock){
	}

