#include <stdio.h>
#include <stdlib.h>
#include "sab82526rx.h"
#include "sab82526sim.h"

#undef RX_PKT_TRACE

using namespace Oscl;

Sab82526RxFifo::Sab82526RxFifo(Sab82526RxFifoOwner& owner):
		_owner(owner),
		_nextOctet(0),
		_length(0),
		_endOfFrame(false)
		{
	}

void	Sab82526RxFifo::reset(){
	_nextOctet	= 0;
	_length		= 0;
	_endOfFrame	= false;
	}

unsigned char	Sab82526RxFifo::read(){
	unsigned char	data;
	data	= _fifo[_nextOctet];
	_nextOctet++;
	_nextOctet	%= _fifoSize;
	return data;
	}

unsigned short	Sab82526RxFifo::getLength(){
	return _length;
	}

// input from line

void	Sab82526RxFifo::put(unsigned char data){
	_fifo[_nextOctet]	= data;
	_nextOctet++;
	_length++;
	if(_nextOctet < _fifoSize) return;
	_nextOctet	= 0;
	_owner.fifoFull(this);
	}

void	Sab82526RxFifo::endOfFrame(){
	_nextOctet	= 0;
	_owner.endOfFrame(this);
	_endOfFrame	= true;
	}

bool	Sab82526RxFifo::isEndOfFrame(){
	return _endOfFrame;
	}

// Sab82526Rx

Sab82526Rx::Sab82526Rx(Sab82526Sim& sab):
		_sab(sab),
		_huntState(*this),
		_mds01Adm0Stage1State(*this),
		_mds01Adm0Stage2State(*this),
		_mds01Adm0Stage3State(*this),
		_mds01Adm0Stage4State(*this),
		_mds01Adm1Stage1State(*this),
		_mds01Adm1Stage2State(*this),
		_state(&_huntState),
		_currentFifo(0),
		_activeFifo(0),
		_rxFifo1(*this),
		_rxFifo2(*this),
		_rxLength(0)
		{
	_rxFifo1.reset();
	_rxFifo2.reset();
	_emptyList.put(&_rxFifo1);
	_emptyList.put(&_rxFifo2);
	_freeList.put(&_delayNode1);
	_freeList.put(&_delayNode2);
	}

void Sab82526Rx::reset(){
	Sab82526RxFifo*	next;
	_rxLength	= 0;
	if(_activeFifo) _activeFifo->reset();
	if(_currentFifo){
		_currentFifo->reset();
		_emptyList.put(_currentFifo);
		_currentFifo	= 0;
		}
	while((next=_pendingList.get())){
		next->reset();
		_emptyList.put(next);
		}
	flushDelayQueue();
	}

void Sab82526Rx::clock(unsigned long bitClocksSinceLastClock){
	}

void Sab82526Rx::enable(){
	}

void Sab82526Rx::disable(){
	}

void Sab82526Rx::enableInterrupt(){
	}

void Sab82526Rx::disableInterrupt(){
	}

unsigned char Sab82526Rx::readByte(unsigned short index){
	if(_currentFifo)
		return _currentFifo->read();
	return 0;
	}

void Sab82526Rx::releaseFifo(){
	if(_currentFifo){
		if(_currentFifo->isEndOfFrame()){
			_rxLength	= 0;
			}
		_currentFifo->reset();
		_emptyList.put(_currentFifo);
		}
	_currentFifo	= _pendingList.get();
	if(_currentFifo){
		if(_currentFifo->isEndOfFrame()){
			_rxLength		+= _currentFifo->getLength();
			_sab.receiveMessageComplete();
			}
		else{
			_rxLength		+= 32;
			_sab.receiveFifoFull();
			}
		}
	}

unsigned short	Sab82526Rx::getLength(){
	return _rxLength;
	}

// FrameReceiver Interface
void Sab82526Rx::receiveFlag(){
#ifdef RX_PKT_TRACE
	fprintf(stderr," Flag\n");
#endif
	_state->flag();
	}

void Sab82526Rx::receive(Octet data){
#ifdef RX_PKT_TRACE
	fprintf(stderr,"%2.2X ",(int)data);
#endif
	_state->receive(data);
	}

// Sab82526RxFifoOwner Interface

void Sab82526Rx::fifoFull(Sab82526RxFifo* fifo){
	if(!_currentFifo){
		_currentFifo	= fifo;
		_rxLength		+= 32;
		_sab.receiveFifoFull();
		}
	else{
		_pendingList.put(fifo);
		}
	_activeFifo	= _emptyList.get();
	}

void Sab82526Rx::endOfFrame(Sab82526RxFifo* fifo){
	if(!_currentFifo){
		_currentFifo	= fifo;
		_rxLength		+= fifo->getLength();
		_sab.receiveMessageComplete();
		}
	else{
		_pendingList.put(fifo);
		}
	_activeFifo	= _emptyList.get();
	}

// RxState state change interface

void Sab82526Rx::setStartOfFrameState(){
	unsigned char	data;
	unsigned char	masked_data;
	data		= _sab._rwShadow.MODE;
	masked_data	= data & 0xC0;
	if(masked_data == 0x40){
		 if(data & 0x20){
			setRxStateMds01Adm1Stage1();
			}
		else{
			setRxStateMds01Adm0Stage1();
			}
		}
	}

void Sab82526Rx::setRxStateHunt(){
	_state	= &_huntState;
	}

void Sab82526Rx::setRxStateMds01Adm0Stage1(){
	_state	= &_mds01Adm0Stage1State;
	}

void Sab82526Rx::setRxStateMds01Adm0Stage2(){
	_state	= &_mds01Adm0Stage2State;
	}

void Sab82526Rx::setRxStateMds01Adm0Stage3(){
	_state	= &_mds01Adm0Stage3State;
	}

void Sab82526Rx::setRxStateMds01Adm0Stage4(){
	_state	= &_mds01Adm0Stage4State;
	}

void Sab82526Rx::setRxStateMds01Adm1Stage1(){
	_state	= &_mds01Adm1Stage1State;
	}

void Sab82526Rx::setRxStateMds01Adm1Stage2(){
	_state	= &_mds01Adm1Stage2State;
	}

// RxState interface

bool				Sab82526Rx::isMyAddressH(Octet data){	// true if data is same as RAH1 or RAH2
	if(data == _sab._writeShadow.RAH1){
		_rsta	= (1<<RSTA_HA1);
		}
	else if(data == _sab._writeShadow.RAH2){
		_rsta	= 0;
		}
	else{
		return false;
		}
	if(data & (0x2)){	// C/R bit
		_rsta	|= (1<<RSTA_CR);
		}
	return true;
	}

bool				Sab82526Rx::isMyAddressL(Octet data){	// true if data is same as RAL1 or RAL2
	if(data == _sab._rwShadow.RAL1){
		_rsta	= (1<<RSTA_LA);
		}
	else if(data == _sab._writeShadow.RAL2){
		}
	else{
		return false;
		}
	return true;
	}

void				Sab82526Rx::resetCRC(){		// reset the CRC accumulator
	// nothing for now (not checking the CRC)
	}

void				Sab82526Rx::accumulateCRC(Octet data){	// accumulate data into current CRC
	// nothing for now (not checking the CRC)
	}

void				Sab82526Rx::runtFrameReceived(){
	// Nothing for Sab82526
	}

Sab82526RxDataNode*	Sab82526Rx::getFreeNode(){
	return _freeList.get();
	}

void				Sab82526Rx::putNodeInDelayQueue(Sab82526RxDataNode* node){
	_delayQ.put(node);
	}

void				Sab82526Rx::frameComplete(){	// check CRC,put RSTA copy in FIFO indicate EOF to FIFO
	// CRC is always good for now
	_rsta	|= (1<<RSTA_VFR) | (1<<RSTA_CRC);
	if(_activeFifo){
		putInActiveFifo(_rsta);
		_activeFifo->endOfFrame();
		}

	flushDelayQueue();
	}

void				Sab82526Rx::flushDelayQueue(){
	Sab82526RxDataNode*	next;
	while((next=_delayQ.get())){
		_freeList.put(next);
		}
	}

Sab82526RxDataNode*	Sab82526Rx::getDelayQueue(){
	return _delayQ.get();
	}

void				Sab82526Rx::putInActiveFifo(Octet data){
	if(!_activeFifo)
		_activeFifo	= _emptyList.get();
	if(_activeFifo)
		_activeFifo->put(data);
	}

void				Sab82526Rx::receiveOverflow(){
	unsigned char	*exir = (unsigned char *)&_sab._readShadow.EXIR;
	*exir	|= (1<<EXIR_RFO);
	}

bool				Sab82526Rx::fifoActive(){
	if(!_activeFifo) _activeFifo	= _emptyList.get();
	return _activeFifo;
	}

// Sab82526RxState

Sab82526RxState::Sab82526RxState(Sab82526Rx& rx):_rx(rx){
	}

void Sab82526RxState::reset(){
	_rx.flushDelayQueue();
	_rx.setRxStateHunt();
	}

// Sab82526RxStateHunt

Sab82526RxStateHunt::Sab82526RxStateHunt(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateHunt::flag(){
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateHunt::receive(Octet data){
	// discard
	};

// Sab82526RxStateMds01Adm0Stage1

Sab82526RxStateMds01Adm0Stage1::Sab82526RxStateMds01Adm0Stage1(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm0Stage1::flag(){
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm0Stage1::receive(Octet data){
	if(_rx.isMyAddressH(data)){
		_rx.resetCRC();
		_rx.accumulateCRC(data);
		_rx.setRxStateMds01Adm0Stage2();
		}
	else{
		_rx.setRxStateHunt();
		}
	}

// Sab82526RxStateMds01Adm0Stage2
Sab82526RxStateMds01Adm0Stage2::Sab82526RxStateMds01Adm0Stage2(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm0Stage2::flag(){
	_rx.runtFrameReceived();
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm0Stage2::receive(Octet data){
	Sab82526RxDataNode*	node;
	node	= _rx.getFreeNode();
	if(!node){
		fprintf(stderr,"No Sab82526RxDataNode unavailble, design error!\n");
		exit(1);
		}
	node->_data	= data;
	_rx.putNodeInDelayQueue(node);
	_rx.accumulateCRC(data);
	_rx.setRxStateMds01Adm0Stage3();
	}

// Sab82526RxStateMds01Adm0Stage3

Sab82526RxStateMds01Adm0Stage3::Sab82526RxStateMds01Adm0Stage3(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm0Stage3::flag(){
	_rx.runtFrameReceived();
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm0Stage3::receive(Octet data){
	Sab82526RxDataNode*	node;
	node	= _rx.getFreeNode();
	if(!node){
		fprintf(stderr,"No Sab82526RxDataNode unavailble, design error!\n");
		exit(1);
		}
	node->_data	= data;
	_rx.putNodeInDelayQueue(node);
	_rx.accumulateCRC(data);
	_rx.setRxStateMds01Adm0Stage4();
	}

// Sab82526RxStateMds01Adm0Stage4

Sab82526RxStateMds01Adm0Stage4::Sab82526RxStateMds01Adm0Stage4(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm0Stage4::flag(){
	if(_rx.fifoActive()){
		_rx.frameComplete();
		}
	_rx.flushDelayQueue();
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm0Stage4::receive(Octet data){
	if(_rx.fifoActive()){
		Sab82526RxDataNode*	node;
		node	= _rx.getDelayQueue();
		if(!node){
			fprintf(stderr,"No Sab82526RxDataNode in delay Q, design error!\n");
			exit(1);
			}
		_rx.putInActiveFifo(node->_data);
		node->_data	= data;
		_rx.putNodeInDelayQueue(node);
		_rx.accumulateCRC(data);
		}
	else{
		_rx.flushDelayQueue();
		_rx.receiveOverflow();
		_rx.setRxStateHunt();
		}
	}

// Sab82526RxStateMds01Adm1Stage1
Sab82526RxStateMds01Adm1Stage1::Sab82526RxStateMds01Adm1Stage1(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm1Stage1::flag(){
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm1Stage1::receive(Octet data){
	if(_rx.isMyAddressH(data)){
		_rx.resetCRC();
		_rx.accumulateCRC(data);
		_rx.setRxStateMds01Adm1Stage2();
		}
	else{
		_rx.setRxStateHunt();
		}
	}

// Sab82526RxStateMds01Adm1Stage2
Sab82526RxStateMds01Adm1Stage2::Sab82526RxStateMds01Adm1Stage2(Sab82526Rx& rx):Sab82526RxState(rx){
	}

void Sab82526RxStateMds01Adm1Stage2::flag(){
	_rx.runtFrameReceived();
	_rx.setStartOfFrameState();
	}

void Sab82526RxStateMds01Adm1Stage2::receive(Octet data){
	if(_rx.isMyAddressL(data)){
		_rx.accumulateCRC(data);
		_rx.setRxStateMds01Adm0Stage2();
		}
	else{
		_rx.setRxStateHunt();
		}
	}

