
/***************************** 68000 SIMULATOR ****************************

File Name: RUN.C
Version: 1.0

This file contains various routines to run a 68000 instruction.  These
routines are :

	decode_size(context,), eff_addr(context,), runprog(), exec_context->inst(), exception()

***************************************************************************/


#define NO_CTRL_C_SIGNAL

#include <stdio.h>
#ifndef NO_CTRL_C_SIGNAL
#include <signal.h>
#endif
#include "sim68k/devices/cif.h"
#include "proto.h"
#include "opcodes.h"        					/* opcode masks for decoding */
														/* the 68000 instructions */



/**************************** int decode_size(context,) ****************************

   name       : int decode_size (context,result)
   parameters : long *result : the appropriate mask for the decoded size
   function   : decodes the size field in the instruction being processed
                  and returns a mask to be used in instruction execution.
                  For example, if the size field was "01" then the mask
                  returned is WORD.

****************************************************************************/


int	decode_size (SimContext *context,long *result)
{
	int	bits;

	/* the size field is always in bits 6 and 7 of the instruction word */
	bits = (context->inst >> 6) & 0x0003;

	switch (bits) {
		case 0 : *result = BYTE;
			 break;
		case 1 : *result = WORD;
			 break;
		case 2 : *result = LONG;
			 break;
		default : *result = 0;
		}

	if (*result != 0)
	   return SUCCESS;
	else
	   return FAILURE;

}



/**************************** int eff_addr(context,) *******************************

   name       : int eff_addr (context,size, mask, add_times)
   parameters : long size : the appropriate size mask
                int mask : the effective address modes mask to be used
                int add_times : tells whether to increment the cycle counter
                      (there are times when we don't want to)
   function   : eff_addr(context,) decodes the effective address field in the current
                  instruction, returns a pointer to the effective address
                  either in context->EA1 or context->EA2, and returns the value at that
                  location in EV1 or EV2. The value is NOT returned if the
                  mask is CONTROL_ADDR.

****************************************************************************/


int eff_addr (SimContext *context,long size, int mask, int add_times)
{
	int						mode;
	int						reg;
	int						legal;
	int						addr;
	int						move_operation;
	int						bwinc;
	int						linc;
	long					ext;
	long					temp_ext;
	long					inc_size;
	long					ind_reg;
	long					value=0;
	long					disp;
	EffectiveAddressType	eaType = effectiveAddressTypeMemoryAddress;

	if (
			(
				((context->inst & 0xf000) == 0x1000)
				|| ((context->inst & 0xf000) == 0x2000)
				||	((context->inst & 0xf000) == 0x3000)
			)
			&& (mask == DATA_ALT_ADDR)
		)
		move_operation = TRUE;
	else
		move_operation = FALSE;

	if (move_operation)
		addr = (context->inst >> 6) & 0x003f;
	else
		addr = context->inst & 0x003f;
	legal = FALSE;
	bwinc = linc = 0;
   if (move_operation)		/* reg and mode are reversed in MOVE context->dest EA */
		{
		reg = (addr & MODE_MASK) >> 3;
		mode = addr & REG_MASK;
		}
	else
		{
		mode = (addr & MODE_MASK) >> 3;
		reg = addr & REG_MASK;
		}
	switch (mode) {
	   case	0 : if (mask & bit_1)
			{
			value = reg;
			bwinc = linc = 0;
			legal = TRUE;
			eaType	= effectiveAddressTypeDataRegister;
			}
		    break;
	   case	1 : if (mask & bit_2)
			{
			reg = a_reg(context,reg);
			value = reg;
			bwinc = linc = 0;
			legal = TRUE;
			eaType	= effectiveAddressTypeAddressRegister;
			}
		    break;
	   case	2 : if (mask & bit_3)
			{
			reg = a_reg(context,reg);
			value = context->A[reg];
			bwinc = 4;
			linc = 8;
			legal = TRUE;
			eaType	= effectiveAddressTypeMemoryAddress;
			}
		    break;
	   case	3 : if (mask & bit_4)
			{
			reg = a_reg(context,reg);
			if (size == BYTE)
				inc_size = 1;
			else if (size == WORD)
				inc_size = 2;
			else inc_size = 4;
			value = context->A[reg];
			context->A[reg] = context->A[reg] + inc_size;
			bwinc = 4;
			linc = 8;
			legal = TRUE;
			eaType	= effectiveAddressTypeMemoryAddress;
			}
		    break;
	   case	4 : if (mask & bit_5)
			{
			reg = a_reg(context,reg);
			if (size == BYTE)
				inc_size = 1;
			else if (size == WORD)
				inc_size = 2;
			else inc_size = 4;
			context->A[reg] = context->A[reg] - inc_size;
			value = context->A[reg];
			bwinc = 6;
			linc = 10;
			legal = TRUE;
			eaType	= effectiveAddressTypeMemoryAddress;
			}
		    break;
	   case	5 : if (mask & bit_6)
			{
			reg = a_reg(context,reg);
			mem_request (context,&context->pc, (long) WORD, &ext);
			from_2s_comp (context,ext, (long) WORD, &ext);
			value = context->A[reg] + ext;
			bwinc = 8;
			linc = 12;
			legal = TRUE;
			eaType	= effectiveAddressTypeMemoryAddress;
			}
		    break;
	   case	6 : if (mask & bit_7)
			{
			int	scale;
			reg = a_reg(context,reg);
			/* fetch extension word */
			mem_request (context,&context->pc, (long) WORD, &ext);
			disp = ext & 0x00ff;
			sign_extend (context,disp, BYTE, &disp);
			from_2s_comp (context,disp, (long) WORD, &disp);
			/* get index register value */
			if (ext & 0x8000)
				ind_reg = context->A[a_reg(context,(ext & 0x7000) >> 12)];
			else
				ind_reg = context->D[(ext & 0x7000) >> 12];

			scale	= (ext & 0x0600) >> 9;
			scale	= 1<<scale;

			/* get correct length for index register */
			if (!(ext & 0x0800))
				{
				sign_extend (context,ind_reg, WORD, &ind_reg);
				from_2s_comp (context,ind_reg, (long) LONG, &ind_reg);
				}
			value = context->A[reg] + disp + (ind_reg*scale);
			bwinc = 10;
			linc = 14;
			legal = TRUE;
			eaType	= effectiveAddressTypeMemoryAddress;
			}
		    break;
	   case	7 : switch (reg) {
			   case	0 : if (mask & bit_8)
	    			{
					mem_request (context,&context->pc, (long) WORD, &ext);
					sign_extend (context,ext, WORD, &ext);
					value = ext;
					bwinc = 8;
					linc = 12;
					legal = TRUE;
					eaType	= effectiveAddressTypeMemoryAddress;
					}
				    break;
			   case	1 : if (mask & bit_9)
					{
					mem_request (context,&context->pc, (long) WORD, &ext);
					mem_request (context,&context->pc, (long) WORD, &temp_ext);

					ext			<<= 16;
					temp_ext	&= 0x0FFFF;
					ext			|= temp_ext;
					value = ext;
					bwinc = 12;
					linc = 16;
					legal = TRUE;
					eaType	= effectiveAddressTypeMemoryAddress;
					}
					 break;
			   case	2 : if (mask & bit_10)
					{
					mem_request (context,&context->pc, (long) WORD, &ext);
					from_2s_comp (context,ext, (long) WORD, &ext);
					value = context->pc + ext - 2;
					bwinc = 8;
					linc = 12;
					legal = TRUE;
					eaType	= effectiveAddressTypeMemoryAddress;
					}
					 break;
			   case	3 : if (mask & bit_11)
					{
					int	scale;
					/* fetch extension word */
					mem_request (context,&context->pc, (long) WORD, &ext);
					disp = ext & 0xff;
					sign_extend (context,disp, BYTE, &disp);
					from_2s_comp (context,disp, (long) WORD, &disp);

					scale	= (ext & 0x0600) >> 9;
					scale	= 1<<scale;

					/* get index register value */
					if (ext & 0x8000)
					   ind_reg = context->A[a_reg(context,(ext & 0x7000) >> 12)];
					else
					   ind_reg = context->D[(ext & 0x7000) >> 12];
					/* get correct length for index register */
					if (!(ext & 0x0800))
						{
						sign_extend (context,ind_reg, WORD, &ind_reg);
						from_2s_comp (context,ind_reg, (long) LONG, &ind_reg);
						}
					ext = ext & 0x00ff;
					/* This kind of instruction indicates an effective address relative to the PC.
					 * The PC is specified to be at the address of the extension word. In this case
					 * the context->pc has been incremented just beyond the extension word. Thus,
					 * it is necessary to subtract 2 from the program counter during the EA calculation.
					 */

					value = context->pc - 2 + (disp & WORD) + (ind_reg*scale);
					bwinc = 10;
					linc = 14;
					legal = TRUE;
					eaType	= effectiveAddressTypeMemoryAddress;
					}
				    break;
			   case	4 : if (mask & bit_12)
					{
					if ((size == BYTE) || (size == WORD))
						mem_request (context,&context->pc, (long) WORD, &ext);
					else
						mem_request (context,&context->pc, LONG, &ext);
					value = ext;
					bwinc = 4;
					linc = 8;
					legal = TRUE;
					eaType	= effectiveAddressTypeImmediate;
					}
				    break;
				}
			break;
	}   	  /* switch */

	if (legal) {
		if (add_times) {
			if (size != LONG) {
				inc_cyc (context,bwinc);
				}
			else {
				inc_cyc (context,linc);
				}
			}
		if (move_operation)/* choose context->EA2 in case of MOVE context->dest effective address */
			{
			context->typeEA2	= eaType;
	 		context->EA2		= value;
			}
		else
			{
			context->typeEA1	= eaType;
			context->EA1		= value;
			}
		return SUCCESS;
		}
	else
		return FAILURE;       /* return FAILURE if illegal addressing mode */

}

#ifndef NO_CTRL_C_SIGNAL
static int	userBreak;

void interruptSignalHandler(int sig)
{
	userBreak	= TRUE;
}
#endif

/**************************** int runprog() *******************************

   name       : int runprog ()
   parameters : NONE
   function   : executes a program at pc specified or current pc if not
                  specified.  this function is the outer loop of the
                  running program.  it handles i/o interrupts, user
                  (keyboard) interrupts, and calls the error routine if an
                  illegal opcode is found.


****************************************************************************/

int runprog(SimContext *context)
{
int		i;

	context->errflg = 0;

	for (;;)		/* execute instructions until error or break */
		{
#ifdef NO_CTRL_C_SIGNAL
		if(processStdin()){	/* if break requested */
#else
		if(userBreak){	/* if ^C sent */
			sigaction(SIGINT,&oact,(struct sigaction *)0);
#endif
			return (USER_BREAK);
			}

		pollDevices(context->cycles);	/* Give time to I/O devices */

		exec_inst(context);		/* execute an instruction */

		if (context->errflg)   /* if illegal opcode in program initiate an exception */
			{
			fprintf(stderr,"illegal opcode at loc %04lx\n", WORD & (context->pc - 1));
			inc_cyc (context,34);
			exception (context,1, 0, READ,context->old_pc,16);
			mem_req (context,0x10, LONG, &context->pc);
#ifndef NO_CTRL_C_SIGNAL
			sigaction(SIGINT,&oact,(struct sigaction *)0);
#endif
			return (0);
			}

		if (context->TRACE) {
			}

		if(context->dataBreakPoint){
			context->dataBreakPoint	= 0;
			return (0);
			}

		for (i = 0; i < context->bpoints; i++)	  /* if pc equals a breakpoint, break */
			if (context->pc == context->brkpt[i])
				{
#ifndef NO_CTRL_C_SIGNAL
				sigaction(SIGINT,&oact,(struct sigaction *)0);
#endif
				return (0);
				}
		if (context->SSTEP) return (0);  /* if single stepping is on then stop running */

		context->old_pc = context->pc;	/* update the context->old_pc */
		}
}




/**************************** int exec_context->inst() *****************************

   name       : int exec_context->inst ()
   parameters : NONE
   function   : executes a single instruction at the location pointed
                  to by pc.  it is called from runprog() and sets the
                  flag "context->errflg" if an illegal opcode is detected so
                  that runprog() terminates.  exec_context->inst() also takes
                  care of handling the different kinds of context->exceptions
                  that may occur.  If an instruction returns a different
                  return code than "SUCCESS" then the appropriate
                  exception is initiated, unless the "context->exceptions" flag
                  is turned off by the user in which case the exception
                  is not initiated and the program simply terminates and
                  informs the user that an exception condition has occurred.


****************************************************************************/


void exec_inst(SimContext *context)
{
int	start, finish, exec_result, i;
long	iinst;
PCTraceRecord*	t;


t	= allocTraceRec(context);
if(!t){
	t	= getTraceRec(context);
	}
if(t){
	t->type		= pcTraceRecordTypeInstruction;
	t->address	= context->pc;
	putTraceRec(context,t);
	}

if ( !(mem_request (context,(long *)&context->pc, (long) WORD, &iinst)) )
   {
	context->inst	= (int)iinst;
	start = offsets[(context->inst & FIRST_FOUR) >> 12];
	finish = offsets[((context->inst & FIRST_FOUR) >> 12) + 1] - 1;
	context->errflg = TRUE;

	if ((context->inst & 0xF000) == 0xA000)	/* unimplemented instruction */
		{
		context->errflg = FALSE;
		mem_req (context,0x28, LONG, &context->pc);
		fprintf (stderr,"'Line A' unimplemented instruction found at location %4lx\n",context->old_pc);
		}
	else
	if ((context->inst & 0xF000) == 0xF000)	/* unimplemented instruction */
		{
		context->errflg = FALSE;
		mem_req (context,0x2C, LONG, &context->pc);
		fprintf (stderr,"'Line F' unimplemented instruction found at location %4lx\n", context->old_pc);
		}
	else
	for (i = start; i <= finish; i++)
		{
		if ((context->inst & inst_arr[i].mask) == inst_arr[i].val)
			{
			exec_result = (*(names[i]))(context);

	if(context->doubleBusFault){
		context->SSTEP = TRUE;	/* break out of the program */
		fprintf (stderr,"Double bus fault.\n");
		fprintf (stderr,"Double bus fault at location %4lx. Execution halted\n", context->old_pc);
		}
	else if (context->exceptions)
	{
	int	ipl;
	int	ipm;
	int	vector;
			switch (exec_result)
				{
				case SUCCESS  : break;
				case BAD_INST : inc_cyc (context,34);	/* load the pc from the illegal */
									 exception (context,1, 0, READ,context->old_pc,16);
									 mem_req (context,0x10, LONG, &context->pc);	    /* vector */
									fprintf (stderr,"Illegal instruction found at location %4lx\n", context->old_pc);
						          break;
				case NO_PRIVILEGE : inc_cyc (context,34);
									 exception (context,1, 0, READ,context->old_pc,32);
									 mem_req (context,0x20, LONG, &context->pc);
									fprintf (stderr,"supervisor privilege violation at location %4lx\n", context->old_pc);
						          break;
				case CHK_EXCEPTION : inc_cyc (context,40);
									 exception (context,2, 0, READ,context->pc,24);
									 mem_req (context,0x18, LONG, &context->pc);
						          break;
				case ILLEGAL_TRAP : inc_cyc (context,34);
									 exception (context,1, 0, READ,context->old_pc,16);
									 mem_req (context,0x10, LONG, &context->pc);
						          break;
				case STOP_TRAP : context->SSTEP = TRUE;	/* break out of the program */
									 break;
				case TRAP_TRAP : inc_cyc (context,34);
									{
									int vectorOffset	= 128 + (context->inst & 0x0f) * 4;
									 exception (context,2, 0, READ,context->pc,vectorOffset);
									 mem_req (context,vectorOffset, LONG, &context->pc);
									}
						          break;
				case TRAPV_TRAP : inc_cyc (context,34);
									 exception (context,2, 0, READ,context->pc,28);
									 mem_req (context,0x1c, LONG, &context->pc);
						          break;
				case DIV_BY_ZERO : inc_cyc (context,38);
									 exception (context,2, 0, READ,context->pc,20);
									 mem_req (context,0x14, LONG, &context->pc);
						          break;
				}

			ipm	= context->sr;
			ipm	&= 0x700;
			ipm	>>= 8;

			if(ipm < 7){	/* interrupts are enabled */
				ipl = interruptPending(&vector);
				if(ipl > ipm){
					inc_cyc (context,34);
					exception (context,2, 0, READ,context->pc,vector*4);
					mem_req (context,vector*4, LONG, &context->pc);
					context->sr	&= ~0x700;
					context->sr	|= ipl<<8;	/* Mask lower priority interrupts */
					}
				}

			if (context->sr & tbit)   /* initiate a TRACE exception */
				{
				inc_cyc (context,34);
				exception (context,1, 0, READ,context->pc,36);
				mem_req (context,0x24, LONG, &context->pc);
				}
	}
	else
	{
			switch (exec_result)
				{
				case SUCCESS  :
					break;
				case BAD_INST :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"Illegal instruction found at location %4lx. Execution halted\n", context->old_pc);
					break;
				case NO_PRIVILEGE :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"supervisor privilege violation at location %4lx. Execution halted\n", context->old_pc);
					break;
				case CHK_EXCEPTION :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"CHK exception occurred at location %4lx. Execution halted\n", context->old_pc);
					break;
				case ILLEGAL_TRAP :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"Illegal instruction found at location %4lx. Execution halted\n", context->old_pc);
					break;
				case STOP_TRAP :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"STOP instruction executed at location %4lx. Execution halted\n", context->old_pc);
					break;
				case TRAP_TRAP :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"TRAP exception occurred at location %4lx. Execution halted\n", context->old_pc);
					break;
				case TRAPV_TRAP :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"TRAPV exception occurred at location %4lx. Execution halted\n", context->old_pc);
					break;
				case DIV_BY_ZERO :
					context->SSTEP = TRUE;	/* break out of the program */
					fprintf (stderr,"Divide by zero occurred at location %4lx. Execution halted\n", context->old_pc);
					break;
				}

	if (context->sr & tbit)
		{
		context->SSTEP = TRUE;	/* break out of the program */
		fprintf (stderr,"TRACE exception occurred at location %4lx. Execution halted\n", context->old_pc);
		}
	}
			context->errflg = FALSE;
			break;
			}
		}

}

}





/**************************** int exception () *****************************

   name       : int exception (class, loc, r_w)
   parameters : int class : class of exception to be taken
                long loc : the address referenced in the case of an
                  address or bus error
                int r_w : in the case of an address or bus error, this
                  tells whether the reference was a read or write
   function   : initiates exception processing by pushing the appropriate
                  exception stack frame on the system stack and turning
                  supervisor mode on and TRACE mode off.


****************************************************************************/

int	exception(SimContext *context,int class, long loc, int r_w, long pc,int vectorOffset)
{
	int	info_word;

	if ( (class == 1) || (class == 2))
		{
		context->A[8] -= 2;
		putLocation (context,context->A[8], (long) vectorOffset, (long) WORD);
		context->A[8] -= 4;		/* create the stack frame for class 1 and 2 context->exceptions */
		putLocation (context,context->A[8], context->pc, LONG);
		context->A[8] -= 2;
		putLocation (context,context->A[8], (long) context->sr, (long) WORD);
		}
	else
		{						/* class 0 exception (address or bus error) */
/*
According to the 68K manual:
	If a bus error occurs during exception processing for a bus error, an
	address error, a reset, or while the processor is loading stack information during RTE
	executtion, the processor halts. ... Only assertion of RESET can restart a halted processor.
*/
		if(context->busOrAddressErrorInProgress){
			context->doubleBusFault	= 1;
			return DOUBLE_FAULT;	/* Stop recursion */
			}
		context->busOrAddressErrorInProgress	= TRUE;

		inc_cyc (context,50);    /* fifty clock context->cycles for the address or bus exception */
		context->A[8] -= 2;
		putLocation (context,context->A[8], (long) 0xC000+vectorOffset, (long) WORD);
		context->A[8] -= 4;						/* now create the exception stack frame */
		putLocation (context,context->A[8], pc, LONG);
		context->A[8] -= 2;
		putLocation (context,context->A[8], (long) context->sr, (long) WORD);
		context->A[8] -= 2;
		putLocation (context,context->A[8], (long) context->inst, (long) WORD);
		context->A[8] -= 4;
		putLocation (context,context->A[8], loc, LONG);
		context->A[8] -= 2;
		info_word = 0x6;
		if (r_w == READ)
			info_word |= 0x10;
		putLocation (context,context->A[8], (long)0x0016, (long) WORD);/* push information word */
		}
	context->sr = context->sr | sbit;			/* force processor into supervisor state */
	context->sr = context->sr & ~tbit;			/* turn off TRACE mode */

return SUCCESS;
}



