
/***************************** 68000 SIMULATOR ****************************

File Name: UTILS.C
Version: 1.0

This file contains various routines to aid in executing a 68000 instruction

The routines are :

   to_2s_comp, from_2s_comp, sign_extend, inc_cyc, eff_addr_code,
   a_reg, mem_put, mem_req, mem_request, put, value_of, cc_update,
   and check_condition.

***************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "proto.h"         /* contains global declarations */
#include "sim68k/devices/cif.h"

static int	io_put (SimContext *context,long data, long loc, long size);
static int	io_get (SimContext *context,long loc, long size, long *result);

inline static int	isMemoryLocation(long loc)
{
	return (( (long)loc >= (long)0 ) && ((long)loc < (long)MEMSIZE))?1:0;
}


/**************************** int to_2s_comp () ****************************

   name       : int to_2s_comp (number, size, context->result)
   parameters : long number : the number to be converted to 2's compliment
                long size : the size of the operation
                long *context->result : the context->result of the conversion
   function   : to_2s_comp() converts a number to 2's compliment notation.


****************************************************************************/

int	to_2s_comp (SimContext *context,long number, long size, long *result)
{

if (size == LONG)
	{
	if (number < 0)
		*result = ~number - 1;
	else
		*result = number;
	}
if (size == WORD)
	{
	if (number < 0)
		*result = (~(number & WORD) & WORD) - 1;
	else
		*result = number;
	}
if (size == BYTE)
	{
	if (number < 0)
		*result = (~(number & BYTE) & BYTE) - 1;
	else
		*result = number;
	}

return SUCCESS;

}




/**************************** int from_2s_comp (context,) **************************

   name       : int from_2s_comp (context,number, size, result)
   parameters : long number : the number to be converted to 2's compliment
                long size : the size of the operation
                long *result : the result of the conversion
   function   : from_2s_comp(context,) converts a number from 2's compliment
                  notation to the "C" language format so that operations
                  may be performed on it.


****************************************************************************/


int	from_2s_comp (SimContext *context,long number, long size, long *result)
{

if (size == LONG)
	{
	if (number & 0x80000000)
		{
		*result = ~number + 1;
		*result = -*result;
		}
	else
		*result = number;
	}
if (size == WORD)
	{
	if (number & 0x8000)
		{
		*result = (~number + 1) & WORD;
		*result = -*result;
		}
	else
		*result = number;
	}
if (size == BYTE)
	{
	if (number & 0x80)
		{
		*result = (~number + 1) & BYTE;
		*result = -*result;
		}
	else
		*result = number;
	}

return SUCCESS;

}



/**************************** int sign_extend (context,) ***************************

   name       : int sign_extend (context,number, size_from, result)
   parameters : int number : the number to sign extended
                long size_from : the size of the source
                long *result : the result of the sign extension
   function   : sign_extend(context,) sign-extends a number from byte to word or
                  from word to long.


****************************************************************************/

int	sign_extend (SimContext *context,int number, long size_from, long *result)
{

	*result = number & size_from;
	if ((size_from == WORD) && (number & 0x8000))
		*result |= 0xffff0000;
	if ((size_from == BYTE) && (number & 0x80))
		*result |= 0xff00;

	return SUCCESS;

}




/**************************** int inc_cyc (context,) *******************************

   name       : int inc_cyc (context,num)
   parameters : int num : number of context->cycles to increment the cycle counter.
   function   : inc_cyc(context,) increments the machine cycle counter by 'num'.


****************************************************************************/

void inc_cyc (SimContext *context,int num)
{

   context->cycles = context->cycles + num;

}



/**************************** int eff_addr_code (context,) *************************

   name       : int eff_addr_code (context,inst, start)
   parameters : int inst : the instruction word
                int start : the start bit of the effective address field
   function   : returns the number of the addressing mode contained in
                  the effective address field of the instruction.  This
                  number is used in calculating the execution time for
                  many functions.


****************************************************************************/

int	eff_addr_code (SimContext *context,int inst, int start)
{
	int	mode, reg;

	inst = (inst >> start);
	reg = inst & 0x07;
	mode = (inst >> 3) & 0x07;

	if (mode != 7) return (mode);

	switch (reg) {
		case 0x00 : return (7);
	               break;
		case 0x01 : return (8);
	               break;
		case 0x02 : return (9);
	               break;
		case 0x03 : return (10);
	               break;
		case 0x04 : return (11);
	               break;
		default   : return (12);
   	            break;
		}

	return -1;

}




/**************************** int a_reg (context,) *********************************

   name       : int a_reg (context,reg_num)
   parameters : int reg_num : the address register number to be processed
   function   : a_reg(context,) allows both the SSP and USP to act as context->A[7].  It
                  returns the value '8' if the supervisor bit is set and
                  the reg_num input was '7'.  Otherwise, it returns the
                  reg_num without change.


****************************************************************************/

int	a_reg (SimContext *context,int reg_num)
{

	if ((reg_num == 7) && (context->sr & sbit))
		return 8;
	else
		return reg_num;

}



/**************************** int mem_put(context,) ********************************

   name       : int mem_put(context,data, loc, size)
   parameters : long data : the data to be placed in memory
                int loc   : the location to place the data
                long size : the appropriate size mask for the operation
   function   : mem_put(context,) puts data in main memory.  It acts as the "memory
                  management unit" in that it checks for out-of-bound
                  virtual addresses and odd memory accesses on long and
                  word operations and performs the appropriate traps in
                  the cases where there is a violation.  Theoretically,
                  this is the only place in the simulator where the main
                  memory should be written to.

****************************************************************************/

int	mem_put (SimContext *context,long data, int loc, long size)
{
PCTraceRecord*	t;
int	err	= 1;

/* check for odd location reference on word and longword writes */
/* if there is a violation, initiate an address exception */
if ((loc % 2 != 0) && (size != BYTE))
	{
	/* generate an address error */
	fprintf (stderr,"Address exception at location %4lx\n", context->old_pc);
	fprintf (stderr,"  was writing to location %4x\n", loc);
	mem_req (context,0xc, LONG, &context->pc);
	exception (context,0, (long) loc, WRITE,context->old_pc,12);
	return (FAILURE);
	}

/* check for virtual-address out-of-bounds condition.  If there is a */
/* violation, initiate a bus error */
if(isMemoryLocation(loc)){
	/* if everything is okay then perform the write according to size */
	if (size == BYTE){
		context->memory[loc] = data & BYTE;
		context->dataBreakPoint=isWrite8Breakpoint(context,loc,(unsigned char)data);
		}
	else if (size == WORD)
		{
		context->memory[loc] = (data >> 8) & BYTE;
		context->memory[loc+1] = data & BYTE;
		context->dataBreakPoint=isWrite16Breakpoint(context,loc,(unsigned short)data);
		}
	else if (size == LONG)
		{
		context->memory[loc] = (data >> 24) & BYTE;
		context->memory[loc+1] = (data >> 16) & BYTE;
		context->memory[loc+2] = (data >> 8) & BYTE;
		context->memory[loc+3] = data & BYTE;
		context->dataBreakPoint=isWrite32Breakpoint(context,loc,(unsigned long)data);
		}
	}
else {
	err	= io_put (context,data,loc,size);
	if(err){/* generate a bus error */
		fprintf (stderr,"Bus Error: Address out of virtual memory space ");
		fprintf (stderr,"at location %4lx\n", context->old_pc);
		fprintf (stderr,"  was writing to location %4x\n", loc);
		mem_req (context,0x8, LONG, &context->pc);
		exception (context,0, (long) loc, WRITE,context->old_pc,8);
		return (FAILURE);
		}
	}

t	= allocTraceRec(context);
if(!t){
	t	= getTraceRec(context);
	}
if(t){
	t->type		= pcTraceRecordTypeWrite;
	t->address	= loc;
	t->data		= data;
	switch(size){
		case BYTE:
			t->size	= pcTraceRecordSizeByte;
			break;
		case WORD:
			t->size	= pcTraceRecordSizeWord;
			break;
		case LONG:
			t->size	= pcTraceRecordSizeLong;
			break;
		}
	putTraceRec(context,t);
	}

return SUCCESS;
}




/**************************** int mem_req(context,) ********************************

   name       : int mem_req(context,loc, size, result)
   parameters : int loc : the memory location to read data from
                long size : the appropriate size mask for the operation
                long *result : a pointer to the longword location
                      to store the result in
   function   : mem_req(context,) returns the contents of a location in main
                  memory.  It acts as the "memory management unit" in
                  that it checks for out-of-bound virtual addresses and
                  odd memory accesses on long and word operations and
                  performs the appropriate traps in the cases where there
                  is a violation.  Theoretically, this is the only function
                  in the simulator where the main memory should be read
                  from.

****************************************************************************/

int	mem_req (SimContext *context,long loc, long size, long *result)
{
PCTraceRecord*	t;
int				err = 1;
long			temp;

/* check for odd location reference on word and longword reads. */
/* If there is a violation, initiate an address exception */
if ((loc % 2 != 0) && (size != BYTE))
	{
	/* generate an address error */
	context->errflg = FALSE;
	fprintf (stderr,"Address exception at location %4lx\n", context->old_pc);
	fprintf (stderr,"  was reading from location %4lx\n", loc);
	mem_req (context,0xc, LONG, &context->pc);
	exception (context,0, (long) loc, READ,context->old_pc,12);
	return (FAILURE);
	}

/* check for virtual-address out-of-bounds condition.  If there is a */
/* violation, initiate a bus error */
if(isMemoryLocation(loc)){
	/* if everything is okay then perform the read according to size */
	temp = context->memory[loc] & BYTE;
	if (size != BYTE)
		{
		temp = (temp << 8) | (context->memory[loc + 1] & BYTE);
		if (size != WORD)
			{
			temp = (temp << 8) | (context->memory[loc + 2] & BYTE);
			temp = (temp << 8) | (context->memory[loc + 3] & BYTE);
			context->dataBreakPoint=isRead32Breakpoint(context,loc,(unsigned long)temp);
			}
		else{
			context->dataBreakPoint=isRead16Breakpoint(context,loc,(unsigned short)temp);
			}
		}
	else{
		context->dataBreakPoint=isRead8Breakpoint(context,loc,(unsigned char)temp);
		}
	*result = temp;
	}
else {
	err = io_get (context,loc,size,result);
	if(err) {
		/* generate a bus error */
		context->errflg = FALSE;
		fprintf (stderr,"Bus Error: Address out of virtual memory space ");
		fprintf (stderr,"at location %4lx\n", context->old_pc);
		fprintf (stderr,"  was reading from location %4lx\n", loc);
		mem_req (context,0x8, LONG, &context->pc);
		exception (context,0, (long) loc, READ,context->old_pc,8);
		return (FAILURE);
		}
	}

t	= allocTraceRec(context);
if(!t){
	t	= getTraceRec(context);
	}
if(t){
	t->type		= pcTraceRecordTypeRead;
	t->address	= loc;
	t->data		= *result;
	switch(size){
		case BYTE:
			t->size	= pcTraceRecordSizeByte;
			break;
		case WORD:
			t->size	= pcTraceRecordSizeWord;
			break;
		case LONG:
			t->size	= pcTraceRecordSizeLong;
			break;
		}
	putTraceRec(context,t);
	}

return SUCCESS;
}




/**************************** int mem_request(context,) ****************************

   name       : int mem_request (context,loc, size, result)
   parameters : int *loc : the memory location to read data from
                long size : the appropriate size mask for the operation
                long *result : a pointer to the longword location
                      to store the result in
   function   : mem_request(context,) is another "level" of main-memory access.
                  It performs the task of calling the functin mem_req(context,)
                  (above) using "WORD" for the size mask if the simulator
                  wants a byte from main memory.  Also, it increments
                  the location pointer passed to it.  This is to
                  facilitate easy opcode and operand fetch operations
                  where the program counter needs to be incremented.

                  Therefore, in the simulator, "mem_req(context,)" requests data
                  from main memory, and "mem_request(context,)" does the same but
                  also increments the location pointer.  Note that
                  mem_request(context,) requires a pointer to an int as the first
                  parameter.

****************************************************************************/

int	mem_request (SimContext *context,long *loc, long size, long *result)
{
	int	req_result;

	if (size == LONG)
		req_result = mem_req (context,*loc, LONG, result);
	else
		req_result = mem_req (context,*loc, (long) WORD, result);

	if (size == BYTE)
		*result = *result & 0xff;

	if (!req_result){
		if (size == LONG){
			*loc += 4;
			}
		else{
			*loc += 2;
			}
		}

	return req_result;

}


#if 0
inline int putDataReg(SimContext* context,int regIndex,long source,long size){
	context->D[regIndex]	= (source & size) | (context->D[regIndex] & ~size);
	return SUCCESS;
	}

inline int putAddressReg(SimContext* context,int regIndex,long source,long size){
	context->A[regIndex]	= (source & size) | (context->A[regIndex] & ~size);
	return SUCCESS;
	}
#endif

int putEA2(SimContext* context,long source,long size){
	switch(context->typeEA2){
		case effectiveAddressTypeDataRegister:
			return putDataReg(context,context->EA2,source,size);
			break;
		case effectiveAddressTypeAddressRegister:
			return putAddressReg(context,context->EA2,source,size);
			break;
		case effectiveAddressTypeMemoryAddress:
			return mem_put (context,source, (int) context->EA2 , size);
			break;
		case effectiveAddressTypeImmediate:
		default:
			/** This is really some other kind of error. **/
			fprintf (stderr,"Simulator error: Attempt to put to immediate effective address.");
			mem_req (context,0x8, LONG, &context->pc);
			exception (context,0, context->EA2, WRITE,context->old_pc,8);
			return (FAILURE);
		}
	}

int putEA1(SimContext* context,long source,long size){
	switch(context->typeEA1){
		case effectiveAddressTypeDataRegister:
			return putDataReg(context,context->EA1,source,size);
			break;
		case effectiveAddressTypeAddressRegister:
			return putAddressReg(context,context->EA1,source,size);
			break;
		case effectiveAddressTypeMemoryAddress:
			return mem_put (context,source, (int) context->EA1, size);
			break;
		case effectiveAddressTypeImmediate:
		default:
			/** This is really some other kind of error. **/
			fprintf (stderr,"Simulator error: Attempt to put to immediate effective address.");
			mem_req (context,0x8, LONG, &context->pc);
			exception (context,0, context->EA1, WRITE,context->old_pc,8);
			return (FAILURE);
		}
	}

#if 0
int putSR(SimContext* context,long source,long size){
	context->sr	= (source & size) | (context->sr & ~size);
	return SUCCESS;
	}

int putResult(SimContext* context,long source,long size){
	context->result	= (source & size) | (context->result & ~size);
	return SUCCESS;
	}

int putLocation(SimContext* context,long address,long source,long size){
	return mem_put (context,source, (int) address, size);
	}

inline int value_of_DataReg(SimContext* context,int regIndex,long* EV,long size){
	*EV	= context->D[regIndex] & size;
	return SUCCESS;
	}

inline int value_of_AddressReg(SimContext* context,int regIndex,long* EV,long size){
	*EV	= context->A[regIndex] & size;
	return SUCCESS;
	}
#endif

int value_of_EA2(SimContext* context,long* EV,long size){
	switch(context->typeEA2){
		case effectiveAddressTypeDataRegister:
			return value_of_DataReg(context,context->EA2,EV,size);
			break;
		case effectiveAddressTypeAddressRegister:
			return value_of_AddressReg(context,context->EA2,EV,size);
			break;
		case effectiveAddressTypeMemoryAddress:
			return mem_req (context, context->EA2, size, EV);
			break;
		case effectiveAddressTypeImmediate:
			*EV	= context->EA2 & size;
			break;
		default:
			fprintf (stderr,"Bus Error: Address out of virtual memory space ");
			fprintf (stderr,"at location %4lx\n", context->old_pc);
			fprintf (stderr,"  was writing to location %4lx\n", (long)context->EA2);
			mem_req (context,0x8, LONG, &context->pc);
			exception (context,0, context->EA2, WRITE,context->old_pc,8);
			return (FAILURE);
		}
	return SUCCESS;
	}

int value_of_EA1(SimContext* context,long* EV,long size){
	switch(context->typeEA1){
		case effectiveAddressTypeDataRegister:
			return value_of_DataReg(context,context->EA1,EV,size);
			break;
		case effectiveAddressTypeAddressRegister:
			return value_of_AddressReg(context,context->EA1,EV,size);
			break;
		case effectiveAddressTypeMemoryAddress:
			return mem_req (context, context->EA1, size, EV);
		case effectiveAddressTypeImmediate:
			*EV	= context->EA1 & size;
			break;
		default:
			fprintf (stderr,"Bus Error: Address out of virtual memory space ");
			fprintf (stderr,"at location %4lx\n", context->old_pc);
			fprintf (stderr,"  was writing to location %4lx\n", (long)context->EA2);
			mem_req (context,0x8, LONG, &context->pc);
			exception (context,0, context->EA1, WRITE,context->old_pc,8);
			return (FAILURE);
		}
	return SUCCESS;
	}

#if 0
int value_of_SR(SimContext* context,long* EV,long size){
	*EV	= context->sr & size;
	return SUCCESS;
	}

int value_of_Location(SimContext* context,long address,long* EV,long size){
	return mem_req (context, address, size, EV);
	}
#endif


/**************************** int cc_update(context,) *****************************

   name       : int cc_update (context,x, n, z, v, c, source, dest, result, size, r)
   parameters : int x, n, z, v, c : the codes for actions that should be
                    taken to compute the different condition codes.
                long source : the source operand for the instruction
                long dest : the destination operand for the instruction
                long result : the result of the instruction
                long size : the size of the instruction
                int r : the shift count for shift and rotate instructions
   function   : updates the five condition codes according to the codes
                  passed as parameters.  each of the condition codes
                  has a number of ways it can be calculated, and the
                  appropriate method of computation is passed as the
                  parameter for that condition code.  The source, dest, and
                  result operands contain the source, destination, and
                  result of the instruction requesting updating the
                  condition codes.  Also, for shift and rotate instructions
                  the shift count needs to be passed.

                  The details of the different ways to calculate condition
                  codes are contained in Appendix A of the 68000
                  Programmer's Reference Manual.

****************************************************************************/


int	cc_update (SimContext *context,int x,int n,int z,int v,int c, long source, long dest, long result, long size,int  r)
{
int		x_bit;
int		n_bit;
int		z_bit;
int		v_bit;
int		c_bit;
long	Rm	= 0;
long	Dm	= 0;
long	Sm	= 0;
long	m	= 0;

/* assign the bits to their variables here */
x_bit = context->sr & xbit;
n_bit = context->sr & nbit;
z_bit = context->sr & zbit;
v_bit = context->sr & vbit;
c_bit = context->sr & cbit;

source &= size;
dest &= size;
result &= size;

if (size == BYTE)
			{
			m = 7;
		   Sm = source & 0x0080;
		   Rm = result & 0x0080;
		   Dm = dest & 0x0080;
			};
if (size == WORD)
			{
			m = 15;
		   Sm = source & 0x8000;
		   Rm = result & 0x8000;
		   Dm = dest & 0x8000;
			};
if (size == LONG)
			{
			m = 31;
		   Sm = source & 0x80000000;
		   Rm = result & 0x80000000;
		   Dm = dest & 0x80000000;
			};

/* calculate each of the five condition codes according to the requested */
/* method of calculation */
switch (n) {
	case N_A : break; 			/* n_bit not affected */
	case GEN : n_bit = (Rm) ? TRUE : FALSE;
		   break;
	case UND : n_bit = !n_bit;		/* undefined result */
		   break;
	}
switch (z) {
	case N_A : break;			/* z_bit not affected */
	case UND : z_bit = !z_bit;		/* z_bit undefined */
		   break;
	case GEN : z_bit = (result == 0) ? TRUE : FALSE;
		   break;
	case CASE_1 : z_bit = (z_bit && (result == 0)) ? TRUE : FALSE;
	      break;
	};
switch (v) {
	case N_A : break;			/* v_bit not affected */
	case ZER : v_bit = 0;
		   break;
	case UND : v_bit = !v_bit;		/* v_bit not defined */
		   break;
	case CASE_1 : v_bit = ((Sm && Dm && !Rm) || (!Sm && !Dm && Rm)) ?
			TRUE : FALSE;
		      break;
	case CASE_2 : v_bit = ((!Sm && Dm && !Rm) || (Sm && !Dm && Rm)) ?
			TRUE : FALSE;
		      break;
	case CASE_3 : v_bit = (Dm && Rm) ? TRUE : FALSE;
		      break;
	};
switch (c) {
	case N_A : break;			/* c_bit not affected */
	case UND : c_bit = !c_bit;    		/* c_bit undefined  */
		   break;
	case ZER : c_bit = 0;
		   break;
	case CASE_1 : c_bit = x_bit;
		      break;
	case CASE_2 : c_bit = ((dest >> (r-1)) & 1) ? TRUE : FALSE;
		      break;
	case CASE_3 : c_bit = ((dest >> (m-r+1)) & 1) ? TRUE : FALSE;
		      break;
	case CASE_4 : c_bit = (Dm || Rm) ? TRUE : FALSE;
		      break;
	case CASE_5 : c_bit = ((Sm && Dm) || (!Rm && Dm) || (Sm && !Rm)) ?
			TRUE : FALSE;
		      break;
	case CASE_6 : c_bit = ((Sm && !Dm) || (Rm && !Dm) || (Sm && Rm)) ?
			TRUE : FALSE;
		      break;
	};
switch (x) {
	case N_A : break;     		/* X condition code not affected */
	case GEN : x_bit = c_bit;		/* general case */
		   break;
	};

/* set context->sr according to results */
context->sr = context->sr & 0xffe0;							/* clear the condition codes */
if (x_bit) context->sr = context->sr | xbit;
if (n_bit) context->sr = context->sr | nbit;
if (z_bit) context->sr = context->sr | zbit;
if (v_bit) context->sr = context->sr | vbit;
if (c_bit) context->sr = context->sr | cbit;

return SUCCESS;

}




/**************************** int check_condition(context,) ************************

   name       : int check_condition (context,condition)
   parameters : int condition : the condition to be checked
   function   : check_condition(context,) checks for the truth of a certain
                  condition and returns the result.  The possible conditions
                  are encoded in DEF.H and can be seen in the switch()
                  statement below.

****************************************************************************/


int	check_condition (SimContext *context,int condition)
{
int	result;

result = FALSE;
switch (condition)
	{
	case T : result = 1;				/* true */
		   break;
	case F : result = 0;				/* false */
		   break;
	case HI : result = !(context->sr & cbit) && !(context->sr & zbit);	/* high */
		   break;
	case LS : result = (context->sr & cbit) || (context->sr & zbit);	/* low or same */
		   break;
	case CC : result = !(context->sr & cbit);		/* carry clear */
		   break;
	case CS : result = (context->sr & cbit);		/* carry set */
		   break;
	case NE : result = !(context->sr & zbit);		/* not equal */
		   break;
	case EQ : result = (context->sr & zbit);		/* equal */
		   break;
	case VC : result = !(context->sr & vbit);		/* overflow clear */
		   break;
	case VS : result = (context->sr & cbit);		/* overflow set */
		   break;
	case PL : result = !(context->sr & nbit);		/* plus */
		   break;
	case MI : result = (context->sr & nbit);                /* minus */
		   break;
	case GE : result = ((context->sr & nbit) && (context->sr & vbit)) ||
          (!(context->sr & nbit) && !(context->sr & vbit));          /* greater or equal */
		   break;
	case LT : result = ((context->sr & nbit) && !(context->sr & vbit))    /* less than */
          || (!(context->sr & nbit) && (context->sr & vbit));
		   break;
	case GT : result = ((context->sr & nbit) && (context->sr & vbit) && !(context->sr & zbit))
          || (!(context->sr & nbit) && !(context->sr & vbit) && !(context->sr & zbit)); /* greater than */
		   break;
	case LE : result = ((context->sr & nbit) && !(context->sr & vbit))    /* less or equal */
          || (!(context->sr & nbit) && (context->sr & vbit)) || (context->sr & zbit);
			break;
	}
return result;

}

static int	io_put (SimContext *context,long data, long loc, long size)
{
int	err = SUCCESS;

	if(size == BYTE){
		err	= writeByte(loc,(unsigned char)data);
		context->dataBreakPoint=isWrite8Breakpoint(context,loc,(unsigned char)data);
		}
	else if(size == WORD){
		err	= writeWord(loc,(unsigned short)data);
		context->dataBreakPoint=isWrite16Breakpoint(context,loc,(unsigned short)data);
		}
	else if(size == LONG){
		err	= writeLong(loc,(unsigned long)data);
		context->dataBreakPoint=isWrite32Breakpoint(context,loc,(unsigned long)data);
		}

	return err;
}

static int	io_get (SimContext *context,long loc, long size, long *result)
{
int	err	= FAILURE;

	if(size == BYTE){
		unsigned char byte;
		err	= readByte(loc,&byte);
		if(!err) *result	= byte;
		context->dataBreakPoint=isRead8Breakpoint(context,loc,(unsigned char)byte);
		}
	else if(size == WORD){
		unsigned short	word;
		err = readWord(loc,&word);
		if(!err) *result	= word;
		context->dataBreakPoint=isRead16Breakpoint(context,loc,(unsigned short)word);
		}
	else if(size == LONG){
		unsigned long	lw;
		err = readLong(loc,&lw);
		if(!err) *result	= lw;
		context->dataBreakPoint=isRead32Breakpoint(context,loc,(unsigned long)lw);
		}

return err;
}

int	/* Bool */ isWrite8Breakpoint(SimContext *context,unsigned long address,unsigned char data){
	SimAccessBreakpoint*	next;
	for(next=context->writeDataBreakList;next;next=next->link){
		if(next->address == address && next->data == data) return TRUE;
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if(next->address == address) return TRUE;
		}
	for(next=context->writeAnyBreakList;next;next=next->link){
		if(next->address == address) return TRUE;
		}
	return FALSE;
	}

int	/* Bool */ isWrite16Breakpoint(SimContext *context,unsigned long address,unsigned short data){
	SimAccessBreakpoint*			next;
	register unsigned long			address_1	= address+1;

	for(next=context->writeDataBreakList;next;next=next->link){
		if(next->address == address){
			if(next->data == data){
				return TRUE;
				}
			}
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if(next->address == address){
			return TRUE;
			}
		else if(next->address == address_1){
			return TRUE;
			}
		}
	for(next=context->writeAnyBreakList;next;next=next->link){
		if(next->address == address){
			return TRUE;
			}
		else if(next->address == address_1){
			return TRUE;
			}
		}
	return FALSE;
	}

int	/* Bool */ isWrite32Breakpoint(SimContext *context,unsigned long address,unsigned long data){
	register SimAccessBreakpoint*	next;
	register unsigned long			address_1	= address+1;
	register unsigned long			address_2	= address+2;
	register unsigned long			address_3	= address+3;

	for(next=context->writeDataBreakList;next;next=next->link){
		if(next->address == address){
			if(next->data == data){
				return TRUE;
				}
			}
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if((address - next->address) < 4){
			if(next->address == address){
				return TRUE;
				}
			else if(next->address == address_1){
				return TRUE;
				}
			else if(next->address == address_2){
				return TRUE;
				}
			else if(next->address == address_3){
				return TRUE;
				}
			}
		}
	for(next=context->writeAnyBreakList;next;next=next->link){
		if((address - next->address) < 4){
			if(next->address == address){
				return TRUE;
				}
			else if(next->address == address_1){
				return TRUE;
				}
			else if(next->address == address_2){
				return TRUE;
				}
			else if(next->address == address_3){
				return TRUE;
				}
			}
		}
	return FALSE;
	}

int	/* Bool */ isRead8Breakpoint(SimContext *context,unsigned long address,unsigned char data){
	SimAccessBreakpoint*	next;
	for(next=context->readDataBreakList;next;next=next->link){
		if((next->address == address) && (next->data == data)) return TRUE;
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if(next->address == address) return TRUE;
		}
	for(next=context->readAnyBreakList;next;next=next->link){
		if(next->address == address) return TRUE;
		}
	return FALSE;
	}

int	/* Bool */ isRead16Breakpoint(SimContext *context,unsigned long address,unsigned short data){
	SimAccessBreakpoint*			next;
	register unsigned long			address_1	= address+1;

	for(next=context->readDataBreakList;next;next=next->link){
		if(next->address == address){
			if(next->data == data){
				return TRUE;
				}
			}
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if(next->address == address){
			return TRUE;
			}
		else if(next->address == address_1){
			return TRUE;
			}
		}
	for(next=context->readAnyBreakList;next;next=next->link){
		if(next->address == address){
			return TRUE;
			}
		else if(next->address == address_1){
			return TRUE;
			}
		}
	return FALSE;
	}

int	/* Bool */ isRead32Breakpoint(SimContext *context,unsigned long address,unsigned long data){
	register SimAccessBreakpoint*	next;
	register unsigned long			address_1	= address+1;
	register unsigned long			address_2	= address+2;
	register unsigned long			address_3	= address+3;

	for(next=context->readDataBreakList;next;next=next->link){
		if(next->address == address){
			if(next->data == data){
				return TRUE;
				}
			}
		}
	for(next=context->rwAnyBreakList;next;next=next->link){
		if((address - next->address) < 4){
			if(next->address == address){
				return TRUE;
				}
			else if(next->address == address_1){
				return TRUE;
				}
			else if(next->address == address_2){
				return TRUE;
				}
			else if(next->address == address_3){
				return TRUE;
				}
			}
		}
	for(next=context->readAnyBreakList;next;next=next->link){
		if((address - next->address) < 4){
			if(next->address == address){
				return TRUE;
				}
			else if(next->address == address_1){
				return TRUE;
				}
			else if(next->address == address_2){
				return TRUE;
				}
			else if(next->address == address_3){
				return TRUE;
				}
			}
		}
	return FALSE;
	}

SimAccessBreakpoint*	newSimAccessBreakpoint(	const char		*name,
												unsigned long	address,
												AccessMode		accessMode,
												unsigned char	data,
												unsigned char	mask
												)
{
SimAccessBreakpoint*	b;

	b	= (SimAccessBreakpoint*)malloc(sizeof(SimAccessBreakpoint));
	if(b){
		b->name	= (char *)malloc(strlen(name)+1);
		if(!b->name){
			free(b);
			return (SimAccessBreakpoint*)0;
			}
		strcpy(b->name,name);
		b->address		= address;
		b->accessMode	= accessMode;
		b->data			= (unsigned char)data;
		b->mask			= (unsigned char)mask;
		}
	return b;
}

void putAccessBreakpoint(SimContext* context,SimAccessBreakpoint* b){
	switch(b->accessMode){
		case accessModeAny:
			b->link						= context->rwAnyBreakList;
			context->rwAnyBreakList		= b;
			break;
		case accessModeRead:
			if(b->mask){
				b->link						= context->readDataBreakList;
				context->readDataBreakList	= b;
				}
			else{
				b->link						= context->readAnyBreakList;
				context->readAnyBreakList	= b;
				}
			break;
		case accessModeWrite:
			if(b->mask){
				b->link						= context->writeDataBreakList;
				context->writeDataBreakList	= b;
				}
			else{
				b->link						= context->writeAnyBreakList;
				context->writeAnyBreakList	= b;
				}
			break;
		}
	}

SimAccessBreakpoint* removeAccessBreakpoint(SimContext* context,const char *name);

void			putTraceRec(SimContext* context,PCTraceRecord* t){
	if(context->head){
		context->tail->link  = t;
		}
	else{
		context->head    = context->tail  = t;
		}
	context->tail    = t;
	t->link  = 0;
	}

PCTraceRecord*	getTraceRec(SimContext* context){
	PCTraceRecord*	next;
	if((next = context->head)){
		context->head	= next->link;
		}
	return(next);
	}

PCTraceRecord*	allocTraceRec(SimContext* context){
	PCTraceRecord*	next;
	next	= context->pool;
	if(next){
		context->pool	= next->link;
		}
	return next;
	}

void			freeTraceRec(SimContext* context,PCTraceRecord* t){
	t->link			= context->pool;
	context->pool	= t;
	}

void			flushTrace(SimContext* context){
	PCTraceRecord*	next;
	while((next=getTraceRec(context))){
		freeTraceRec(context,next);
		}
	}

